import copy
import collections
import logging
import os
import sys

import matplotlib.pyplot as plt
import numpy
import statistics

import pyproteininference
from pyproteininference.inference import Inference
from pyproteininference.pipeline import ProteinInferencePipeline

logger = logging.getLogger(__name__)

# set up our logger
logging.basicConfig(
    stream=sys.stderr,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class HeuristicPipeline(ProteinInferencePipeline):
    """
    This is the Protein Inference Heuristic class which houses the logic to run the Protein Inference Heuristic method
     to determine the best inference method for the given data.
    Logic is executed in the [execute][pyproteininference.heuristic.HeuristicPipeline.execute] method.

    Attributes:
        parameter_file (str): Path to Protein Inference Yaml Parameter File.
        database_file (str): Path to Fasta database used in proteomics search.
        target_files (str/list): Path to Target Psm File (Or a list of files).
        decoy_files (str/list): Path to Decoy Psm File (Or a list of files).
        combined_files (str/list): Path to Combined Psm File (Or a list of files).
        target_directory (str): Path to Directory containing Target Psm Files.
        decoy_directory (str): Path to Directory containing Decoy Psm Files.
        combined_directory (str): Path to Directory containing Combined Psm Files.
        output_directory (str): Path to Directory where output will be written.
        output_filename (str): Path to Filename where output will be written. Will override output_directory.
        id_splitting (bool): True/False on whether to split protein IDs in the digest.
            Advanced usage only.
        append_alt_from_db (bool): True/False on whether to append
            alternative proteins from the DB digestion in Reader class.
        pdf_filename (str): Filepath to be written to by Heuristic Plotting method.
            This is optional and a default filename will be created in output_directory if this is left as None.
        inference_method_list (list): List of inference methods used in heuristic determination.
        datastore_dict (dict): Dictionary of [DataStore][pyproteininference.datastore.DataStore]
            objects generated in heuristic determination with the inference method as the key of each entry.
        selected_methods (list): a list of String representations of the selected inference methods based on the
            heuristic.
        selected_datastores (dict):
            a Dictionary of [DataStore object][pyproteininference.datastore.DataStore] objects as selected by the
            heuristic.
        output_type (str): How to output results. Can either be "all" or "optimal". Will either output all results
            or will only output the optimal results.

    """

    RATIO_CONSTANT = 2
    OUTPUT_TYPES = ["all", "optimal"]

    def __init__(
        self,
        parameter_file=None,
        database_file=None,
        target_files=None,
        decoy_files=None,
        combined_files=None,
        target_directory=None,
        decoy_directory=None,
        combined_directory=None,
        output_directory=None,
        output_filename=None,
        id_splitting=False,
        append_alt_from_db=True,
        pdf_filename=None,
        output_type="all",
    ):
        """

        Args:
            parameter_file (str): Path to Protein Inference Yaml Parameter File.
            database_file (str): Path to Fasta database used in proteomics search.
            target_files (str/list): Path to Target Psm File (Or a list of files).
            decoy_files (str/list): Path to Decoy Psm File (Or a list of files).
            combined_files (str/list): Path to Combined Psm File (Or a list of files).
            target_directory (str): Path to Directory containing Target Psm Files.
            decoy_directory (str): Path to Directory containing Decoy Psm Files.
            combined_directory (str): Path to Directory containing Combined Psm Files.
            output_directory (str): Path to Directory where output will be written.
            output_filename (str): Path to Filename where output will be written.
                Will override output_directory.
            id_splitting (bool): True/False on whether to split protein IDs in the digest.
                Advanced usage only.
            append_alt_from_db (bool): True/False on whether to append alternative proteins
                from the DB digestion in Reader class.
            pdf_filename (str): Filepath to be written to by Heuristic Plotting method.
                This is optional and a default filename will be created in output_directory if this is left as None
            output_type (str): How to output results. Can either be "all" or "optimal". Will either output all results
                        or will only output the optimal results.

        Returns:
            HeuristicPipeline: [HeuristicPipeline][pyproteininference.heuristic.HeuristicPipeline] object

        Example:
            >>> heuristic = pyproteininference.heuristic.HeuristicPipeline(
            >>>     parameter_file=yaml_params,
            >>>     database_file=database,
            >>>     target_files=target,
            >>>     decoy_files=decoy,
            >>>     combined_files=combined_files,
            >>>     target_directory=target_directory,
            >>>     decoy_directory=decoy_directory,
            >>>     combined_directory=combined_directory,
            >>>     output_directory=dir_name,
            >>>     output_filename=output_filename,
            >>>     append_alt_from_db=append_alt,
            >>>     pdf_filename=pdf_filename,
            >>>     output_type="all"
            >>> )
        """

        self.parameter_file = parameter_file
        self.database_file = database_file
        self.target_files = target_files
        self.decoy_files = decoy_files
        self.combined_files = combined_files
        self.target_directory = target_directory
        self.decoy_directory = decoy_directory
        self.combined_directory = combined_directory
        self.output_directory = output_directory
        self.output_filename = output_filename
        self.id_splitting = id_splitting
        self.append_alt_from_db = append_alt_from_db
        self.output_type = output_type
        if self.output_type not in self.OUTPUT_TYPES:
            raise ValueError("The variable output_type must be set to either 'all' or 'optimal'")
        if not pdf_filename:
            if self.output_directory and not self.output_filename:
                self.pdf_filename = os.path.join(self.output_directory, "heuristic_plot.pdf")
            elif self.output_filename:
                self.pdf_filename = os.path.join(os.path.split(self.output_filename)[0], "heuristic_plot.pdf")
            else:
                self.pdf_filename = os.path.join(os.getcwd(), "heuristic_plot.pdf")

        else:
            self.pdf_filename = pdf_filename

        self.inference_method_list = [
            Inference.INCLUSION,
            Inference.EXCLUSION,
            Inference.PARSIMONY,
            Inference.PEPTIDE_CENTRIC,
        ]
        self.datastore_dict = {}
        self.selected_methods = None
        self.selected_datastores = {}

        self._validate_input()

        self._set_output_directory()

        self._log_append_alt_from_db()

    def execute(self, fdr_threshold=0.05):
        """
        This method is the main driver of the heuristic method.
        This method calls other classes and methods that make up the heuristic pipeline.
        This includes but is not limited to:

        1. Loops over the main inference methods: Inclusion, Exclusion, Parsimony, and Peptide Centric.
        2. Determines the optimal inference method based on the input data as well as the database file.
        3. Outputs the results and indicates the optimal results.

        Args:
            fdr_threshold (float): The Qvalue/FDR threshold the heuristic method uses to base calculations from.

        Returns:
            None:

        Example:
            >>> heuristic = pyproteininference.heuristic.HeuristicPipeline(
            >>>     parameter_file=yaml_params,
            >>>     database_file=database,
            >>>     target_files=target,
            >>>     decoy_files=decoy,
            >>>     combined_files=combined_files,
            >>>     target_directory=target_directory,
            >>>     decoy_directory=decoy_directory,
            >>>     combined_directory=combined_directory,
            >>>     output_directory=dir_name,
            >>>     output_filename=output_filename,
            >>>     append_alt_from_db=append_alt,
            >>>     pdf_filename=pdf_filename,
            >>>     output_type="all"
            >>> )
            >>> heuristic.execute(fdr_threshold=0.05)

        """

        pyproteininference_parameters = pyproteininference.parameters.ProteinInferenceParameter(
            yaml_param_filepath=self.parameter_file
        )

        digest = pyproteininference.in_silico_digest.PyteomicsDigest(
            database_path=self.database_file,
            digest_type=pyproteininference_parameters.digest_type,
            missed_cleavages=pyproteininference_parameters.missed_cleavages,
            reviewed_identifier_symbol=pyproteininference_parameters.reviewed_identifier_symbol,
            max_peptide_length=pyproteininference_parameters.restrict_peptide_length,
            id_splitting=self.id_splitting,
        )
        if self.database_file:
            logger.info("Running In Silico Database Digest on file {}".format(self.database_file))
            digest.digest_fasta_database()
        else:
            logger.warning(
                "No Database File provided, Skipping database digest and only taking protein-peptide mapping from the "
                "input files."
            )

        for inference_method in self.inference_method_list:

            method_specific_parameters = copy.deepcopy(pyproteininference_parameters)

            logger.info("Overriding inference type {}".format(method_specific_parameters.inference_type))

            method_specific_parameters.inference_type = inference_method

            logger.info("New inference type {}".format(method_specific_parameters.inference_type))
            logger.info("FDR Threshold Set to {}".format(method_specific_parameters.fdr))

            reader = pyproteininference.reader.GenericReader(
                target_file=self.target_files,
                decoy_file=self.decoy_files,
                combined_files=self.combined_files,
                parameter_file_object=method_specific_parameters,
                digest=digest,
                append_alt_from_db=self.append_alt_from_db,
            )
            reader.read_psms()

            data = pyproteininference.datastore.DataStore(reader=reader, digest=digest)

            data.restrict_psm_data()

            data.recover_mapping()

            data.create_scoring_input()

            if method_specific_parameters.inference_type == Inference.EXCLUSION:
                data.exclude_non_distinguishing_peptides()

            score = pyproteininference.scoring.Score(data=data)
            score.score_psms(score_method=method_specific_parameters.protein_score)

            if method_specific_parameters.picker:
                data.protein_picker()
            else:
                pass

            pyproteininference.inference.Inference.run_inference(data=data, digest=digest)

            data.calculate_q_values()

            self.datastore_dict[inference_method] = data

        self.selected_methods = self.determine_optimal_inference_method(
            false_discovery_rate_threshold=fdr_threshold, pdf_filename=self.pdf_filename
        )
        self.selected_datastores = {x: self.datastore_dict[x] for x in self.selected_methods}

        if self.output_type == "all":
            self._write_all_results(parameters=method_specific_parameters)
        elif self.output_type == "optimal":
            self._write_optimal_results(parameters=method_specific_parameters)
        else:
            self._write_optimal_results(parameters=method_specific_parameters)

    def generate_roc_plot(self, fdr_max=0.2, pdf_filename=None):
        """
        This method produces a PDF ROC plot overlaying the 4 inference methods apart of the heuristic algorithm.

        Args:
            fdr_max (float): Max FDR to display on the plot.
            pdf_filename (str): Filename to write roc plot to.

        Returns:
            None:

        """
        f = plt.figure()
        for inference_method in self.datastore_dict.keys():
            fdr_vs_target_hits = self.datastore_dict[inference_method].generate_fdr_vs_target_hits(fdr_max=fdr_max)
            fdrs = [x[0] for x in fdr_vs_target_hits]
            target_hits = [x[1] for x in fdr_vs_target_hits]
            plt.plot(fdrs, target_hits, '-', label=inference_method.replace("_", " "))
            target_fdr = self.datastore_dict[inference_method].parameter_file_object.fdr
            if inference_method in self.selected_methods:
                best_value = min(fdrs, key=lambda x: abs(x - target_fdr))
                best_index = fdrs.index(best_value)
                best_target_hit_value = target_hits[best_index]  # noqa F841

        plt.axvline(target_fdr, color="black", linestyle='--', alpha=0.75, label="Target FDR")
        plt.legend()
        plt.xlabel('Decoy FDR')
        plt.ylabel('Target Protein Hits')
        plt.xlim([-0.01, fdr_max])
        plt.legend(loc='lower right')
        plt.title("FDR vs Target Protein Hits per Inference Method")
        if pdf_filename:
            logger.info("Writing ROC plot to: {}".format(pdf_filename))
            f.savefig(pdf_filename)
        plt.close()

    def _write_all_results(self, parameters):
        """
        Internal method that loops over all results and writes them out.
        """
        for method in list(self.datastore_dict.keys()):
            datastore = self.datastore_dict[method]
            if method in self.selected_methods:
                inference_method_string = "{}_{}".format(method, "optimal_method")
            else:
                inference_method_string = method
            if not self.output_filename and self.output_directory:
                # If a filename is not provided then construct one using output_directory
                # Note: output_directory will always get set even if its set as None - gets set to cwd
                inference_filename = os.path.join(
                    self.output_directory,
                    "{}_{}_{}_{}_{}".format(
                        inference_method_string,
                        parameters.tag,
                        datastore.short_protein_score,
                        datastore.psm_score,
                        "protein_inference_results.csv",
                    ),
                )
            if self.output_filename:
                # If the user specified an output filename then split it apart and insert the inference method
                # Then reconstruct the file
                split = os.path.split(self.output_filename)
                path = split[0]
                filename = split[1]
                inference_filename = os.path.join(path, "{}_{}".format(inference_method_string, filename))
            export = pyproteininference.export.Export(data=self.datastore_dict[method])
            export.export_to_csv(
                output_filename=inference_filename,
                directory=self.output_directory,
                export_type=parameters.export,
            )

    def _write_optimal_results(self, parameters):
        """
        Internal method that writes out the optimized results.
        """

        for method in self.selected_methods:
            datastore = self.datastore_dict[method]
            inference_method_string = "{}_{}".format(method, "optimal_method")
            if not self.output_filename and self.output_directory:
                # If a filename is not provided then construct one using output_directory
                # Note: output_directory will always get set even if its set as None - gets set to cwd
                inference_filename = os.path.join(
                    self.output_directory,
                    "{}_{}_{}_{}_{}".format(
                        inference_method_string,
                        parameters.tag,
                        datastore.short_protein_score,
                        datastore.psm_score,
                        "protein_inference_results.csv",
                    ),
                )
            if self.output_filename:
                # If the user specified an output filename then split it apart and insert the inference method
                # Then reconstruct the file
                split = os.path.split(self.output_filename)
                path = split[0]
                filename = split[1]
                inference_filename = os.path.join(path, "{}_{}".format(inference_method_string, filename))
            export = pyproteininference.export.Export(data=self.selected_datastores[method])
            export.export_to_csv(
                output_filename=inference_filename,
                directory=self.output_directory,
                export_type=parameters.export,
            )

    def determine_optimal_inference_method(
        self,
        false_discovery_rate_threshold=0.05,
        upper_empirical_threshold=1,
        lower_empirical_threshold=0.5,
        pdf_filename=None,
    ):
        """
        This method determines the optimal inference method from Inclusion, Exclusion, Parsimony, Peptide-Centric.

        Args:
            false_discovery_rate_threshold (float): The fdr threshold to use in heuristic algorithm -
                This parameter determines the maximum fdr used when creating a range of finite FDR values.
            upper_empirical_threshold (float): Upper Threshold used for parsimony/peptide centric cutoff for
                the heuristic algorithm.
            lower_empirical_threshold (float): Lower Threshold used for inclusion/exclusion cutoff for
                the heuristic algorithm.
            pdf_filename (str): Filename to write heuristic density plot to.


        Returns:
            list: List of string representations of the recommended inference methods.

        """

        # Get the number of passing proteins
        number_stdev_from_mean_dict = {}
        fdrs = [false_discovery_rate_threshold * 0.01 * x for x in range(100)]
        for fdr in fdrs:
            stdev_from_mean = self.determine_number_stdev_from_mean(false_discovery_rate=fdr)
            number_stdev_from_mean_dict[fdr] = stdev_from_mean

        stdev_collection = collections.defaultdict(list)
        for fdr in fdrs:
            for key in number_stdev_from_mean_dict[fdr]:
                stdev_collection[key].append(number_stdev_from_mean_dict[fdr][key])

        heuristic_scores = self.generate_density_plot(
            number_stdevs_from_mean=stdev_collection, pdf_filename=pdf_filename
        )

        # Apply conditional statement with lower and upper thresholds
        if (
            heuristic_scores[Inference.PARSIMONY] <= lower_empirical_threshold
            or heuristic_scores[Inference.PEPTIDE_CENTRIC] <= lower_empirical_threshold
        ):
            # If parsimony or peptide centric are less than the lower empirical threshold
            # Then select the best method of the two
            logger.info(
                "Either parsimony {} or peptide centric {} pass empirical threshold {}. "
                "Selecting the best method of the two.".format(
                    heuristic_scores[Inference.PARSIMONY],
                    heuristic_scores[Inference.PEPTIDE_CENTRIC],
                    lower_empirical_threshold,
                )
            )
            sub_dict = {
                Inference.PARSIMONY: heuristic_scores[Inference.PARSIMONY],
                Inference.PEPTIDE_CENTRIC: heuristic_scores[Inference.PEPTIDE_CENTRIC],
            }

            if (
                heuristic_scores[Inference.PARSIMONY] <= lower_empirical_threshold
                and heuristic_scores[Inference.PEPTIDE_CENTRIC] <= lower_empirical_threshold
            ):
                # If both are under the threshold return both
                selected_methods = [Inference.PARSIMONY, Inference.PEPTIDE_CENTRIC]

            else:
                selected_methods = [min(sub_dict, key=sub_dict.get)]

        # If the above condition does not apply
        elif (
            heuristic_scores[Inference.EXCLUSION] <= upper_empirical_threshold
            or heuristic_scores[Inference.INCLUSION] <= upper_empirical_threshold
        ):
            # If exclusion or inclusion are less than the upper empirical threshold
            # Then select the best method of the two
            logger.info(
                "Either inclusion {} or exclusion {} pass empirical threshold {}. "
                "Selecting the best method of the two.".format(
                    heuristic_scores[Inference.INCLUSION],
                    heuristic_scores[Inference.EXCLUSION],
                    upper_empirical_threshold,
                )
            )
            sub_dict = {
                Inference.EXCLUSION: heuristic_scores[Inference.EXCLUSION],
                Inference.INCLUSION: heuristic_scores[Inference.INCLUSION],
            }

            if (
                heuristic_scores[Inference.EXCLUSION] <= upper_empirical_threshold
                and heuristic_scores[Inference.INCLUSION] <= upper_empirical_threshold
            ):
                # If both are under the threshold return both
                selected_methods = [Inference.INCLUSION, Inference.EXCLUSION]

            else:
                selected_methods = [min(sub_dict, key=sub_dict.get)]

        else:
            # If we have no conditional scenarios...
            # Select the best method
            logger.info("No methods pass empirical thresholds, selecting the best method")
            selected_methods = [min(heuristic_scores, key=heuristic_scores.get)]

        logger.info("Method(s) {} selected with the heuristic algorithm".format(", ".join(selected_methods)))
        return selected_methods

    def generate_density_plot(self, number_stdevs_from_mean, pdf_filename=None):
        """
        This method produces a PDF Density Plot plot overlaying the 4 inference methods part of the heuristic algorithm.

        Args:
            number_stdevs_from_mean (dict): a dictionary of the number of standard deviations from the mean per
                inference method for a range of FDRs.
            pdf_filename (str): Filename to write heuristic density plot to.

        Returns:
            dict: a dictionary of heuristic scores per inference method which correlates to the
                maximum point of the density plot per inference method.

        """
        f = plt.figure()

        heuristic_scores = {}
        for method in number_stdevs_from_mean:
            readible_method_name = Inference.INFERENCE_NAME_MAP[method]
            kwargs = dict(histtype='stepfilled', alpha=0.3, density=True, bins=40, ec="k", label=readible_method_name)
            x, y, _ = plt.hist(number_stdevs_from_mean[method], **kwargs)
            center = y[list(x).index(max(x))]
            heuristic_scores[method] = abs(center)

        plt.axvline(0, color="black", linestyle='--', alpha=0.75)
        plt.title("Density Plot of the Number of Standard Deviations from the Mean")
        plt.xlabel('Number of Standard Deviations from the Mean')
        plt.ylabel('Number of Observations')
        plt.legend(loc='upper right')
        if pdf_filename:
            logger.info("Writing Heuristic Density plot to: {}".format(pdf_filename))
            f.savefig(pdf_filename)
        else:
            plt.show()
        plt.close()

        logger.info("Heuristic Scores")
        logger.info(heuristic_scores)

        return heuristic_scores

    def determine_number_stdev_from_mean(self, false_discovery_rate):
        """
        This method calculates the mean of the number of proteins identified at a specific FDR of all
        4 methods and then for each method calculates the number of standard deviations
        from the previous calculated mean.

        Args:
            false_discovery_rate (float): The false discovery rate used as a cutoff for calculations.

        Returns:
            dict: a dictionary of the number of standard deviations away from the mean per inference method.

        """

        filtered_protein_objects = {
            x: self.datastore_dict[x].get_protein_objects(
                fdr_restricted=True, false_discovery_rate=false_discovery_rate
            )
            for x in self.datastore_dict.keys()
        }
        number_passing_proteins = {x: len(filtered_protein_objects[x]) for x in filtered_protein_objects.keys()}

        # Calculate how similar the number of passing proteins is for each method
        all_values = [x for x in number_passing_proteins.values()]
        mean = numpy.mean(all_values)
        standard_deviation = statistics.stdev(all_values)
        number_stdev_from_mean_dict = {}
        for key in number_passing_proteins.keys():
            cur_value = number_passing_proteins[key]
            number_stdev_from_mean_dict[key] = (cur_value - mean) / standard_deviation

        return number_stdev_from_mean_dict



#!/usr/bin/python

import argparse

import pyproteininference


def main():
    """
    Script function for running the execute method of the ProteinInferencePipeline class.

    """
    parser = argparse.ArgumentParser(description="Protein Inference")
    parser.add_argument(
        "-t",
        "--target",
        dest="target",
        required=False,
        help="Input target psm output from percolator. Can either input one file or a list of files.",
        metavar="FILE",
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "-d",
        "--decoy",
        dest="decoy",
        required=False,
        help="Input decoy psm output from percolator. Can either input one file or a list of files.",
        metavar="FILE",
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "-f",
        "--combined_files",
        dest="combined_files",
        required=False,
        help="Input combined psm output from percolator. This should contain Target and Decoy PSMS. "
        "Can either input one file or a list of files.",
        metavar="FILE",
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="dir_name",
        required=False,
        help="Result Directory to write to - the name of file will be determined by parameters selected "
        "and tag parameter. If this option is not set, will write results to current working directory.",
        metavar="DIR",
    )
    parser.add_argument(
        "-l",
        "--output_filename",
        dest="output_filename",
        required=False,
        help="Filename to write results to. Can be left blank. If this flag is left blank the filename will be "
        "automatically generated. If set this flag will override -o.",
        metavar="FILE",
        default=None,
    )
    parser.add_argument(
        "-a",
        "--target_directory",
        dest="target_directory",
        required=False,
        help="Directory that contains either .txt or .tsv input target psm data. Make sure the directory ONLY contains"
        " result files.",
        metavar="DIR",
    )
    parser.add_argument(
        "-b",
        "--decoy_directory",
        dest="decoy_directory",
        required=False,
        help="Directory that contains either .txt or .tsv input decoy psm data. Make sure the directory ONLY contains"
        " result files.",
        metavar="DIR",
    )
    parser.add_argument(
        "-c",
        "--combined_directory",
        dest="combined_directory",
        required=False,
        help="Directory that contains either .txt or .tsv input data with targets/decoys combined."
        " Make sure the directory ONLY contains result files.",
        metavar="DIR",
    )
    parser.add_argument(
        "-db",
        "--database",
        dest="database",
        required=False,
        help="Path to the fasta formatted database used in the MS search. This is optional. If not set, "
        "will use the proteins only in the input files.",
        metavar="FILE",
    )
    parser.add_argument(
        "-y",
        "--yaml_params",
        dest="yaml_params",
        required=False,
        help="Path to a Protein Inference Yaml Parameter File. If this is not set, default parameters will be used.",
        metavar="FILE",
    )
    parser.add_argument(
        "-p",
        '--skip_append_alt',
        action="store_true",
        default=False,
        help="Advanced usage only. If this flag is set, will skip adding alternative proteins to each PSM "
        "from the database digest. If this flag is not set, the peptide/protein mapping will be taken "
        "from database digest and appended to the mapping present in the input files.",
    )
    parser.add_argument(
        "-i",
        '--id_splitting',
        action="store_true",
        default=False,
        help="Advanced usage only. If set this flag will split protein identifiers."
        "If not set, this flag will not split protein identifiers."
        "Sometimes the fasta database protein IDs are formatted as: 'sp|ARAF_HUMAN|P10398'. "
        "While protein IDs in the input files are formatted as 'ARAF_HUMAN|P10398'. "
        "Setting This flag will split off the front 'sp|' or 'tr|' from the database protein identifiers. ",
    )

    args = parser.parse_args()

    pipeline = pyproteininference.pipeline.ProteinInferencePipeline(
        parameter_file=args.yaml_params,
        database_file=args.database,
        target_files=args.target,
        decoy_files=args.decoy,
        combined_files=args.combined_files,
        target_directory=args.target_directory,
        decoy_directory=args.decoy_directory,
        combined_directory=args.combined_directory,
        output_directory=args.dir_name,
        output_filename=args.output_filename,
        append_alt_from_db=not args.skip_append_alt,  # Need to reverse the Boolean here
        id_splitting=args.id_splitting,
    )
    pipeline.execute()


if __name__ == "__main__":
    main()



import os
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_parsimony_pulp.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')
for sub_dir in ["leads", "all", "peptides", "psms", "psm_ids"]:
    if not os.path.exists(os.path.join(OUTPUT_DIR, sub_dir)):
        os.makedirs(os.path.join(OUTPUT_DIR, sub_dir))


IDENTIFIER_INDEX = 0
SCORE_INDEX = 1
Q_VALUE_INDEX = 2
GROUP_ID_INDEX = 5
PEPTIDES_INDEX = 6


class TestLoadParsimonyPulpWorkflowSharedPeptideReassignment(TestCase):
    def test_workflow_parsimony_pulp_shared_peptides_best(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # Set the shared_peptides param to "best"
        protein_inference_parameters.shared_peptides = "best"

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        self.assertEqual(len(data.main_data_restricted), 26)

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method=protein_inference_parameters.protein_score)

        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        if protein_inference_parameters.picker:
            data.protein_picker()
        else:
            pass

        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        inference_type = protein_inference_parameters.inference_type

        if inference_type == pyproteininference.inference.Inference.PARSIMONY:
            group = pyproteininference.inference.Parsimony(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.INCLUSION:
            group = pyproteininference.inference.Inclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.EXCLUSION:
            group = pyproteininference.inference.Exclusion(data=data, digest=digest)
            group.infer_proteins()

        lead_protein_psms = [x[0].psms for x in data.grouped_scored_proteins]

        # Get the PSM Identifiers from our lead proteins
        psm_id_list = []
        for protein in lead_protein_psms:
            for psm in protein:
                psm_id_list.append(psm.identifier)

        # Make sure the there are no duplicate psms in our lead proteins
        self.assertEqual(len(psm_id_list), len(set(psm_id_list)))

    def test_workflow_parsimony_pulp_shared_peptides_all(self):

        # Do the same analysis but put shared peptides everywhere

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters_all = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # Set the shared_peptides param to "best"
        protein_inference_parameters_all.shared_peptides = "all"

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest_all = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters_all.digest_type,
            missed_cleavages=protein_inference_parameters_all.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters_all.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters_all.restrict_peptide_length,
            id_splitting=True,
        )
        digest_all.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data_all = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters_all,
            digest=digest_all,
            append_alt_from_db=False,
        )
        pep_and_prot_data_all.read_psms()

        self.assertEqual(len(pep_and_prot_data_all.psms), 27)

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data_all = pyproteininference.datastore.DataStore(pep_and_prot_data_all, digest=digest_all)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data_all.restrict_psm_data()

        self.assertEqual(len(data_all.main_data_restricted), 26)

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data_all.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters_all.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data_all.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score_all = pyproteininference.scoring.Score(data=data_all)
        score_all.score_psms(score_method=protein_inference_parameters_all.protein_score)

        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        if protein_inference_parameters_all.picker:
            data_all.protein_picker()
        else:
            pass

        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        inference_type = protein_inference_parameters_all.inference_type

        if inference_type == pyproteininference.inference.Inference.PARSIMONY:
            group = pyproteininference.inference.Parsimony(data=data_all, digest=digest_all)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.INCLUSION:
            group = pyproteininference.inference.Inclusion(data=data_all, digest=digest_all)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.EXCLUSION:
            group = pyproteininference.inference.Exclusion(data=data_all, digest=digest_all)
            group.infer_proteins()

        lead_protein_psms = [x[0].psms for x in data_all.grouped_scored_proteins]

        # Get the PSM Identifiers from our lead proteins
        psm_id_list = []
        for protein in lead_protein_psms:
            for psm in protein:
                psm_id_list.append(psm.identifier)

        # Make sure the there are duplicate psms in our lead protein objects
        self.assertNotEqual(len(psm_id_list), len(set(psm_id_list)))



from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")

TARGET_FILE_MULTIPLICATIVE = resource_filename(
    "pyproteininference", "../tests/data/test_perc_data_target_multiplicative.txt"
)
DECOY_FILE_MULTIPLICATIVE = resource_filename(
    "pyproteininference", "../tests/data/test_perc_data_decoy_multiplicative.txt"
)
PARAMETER_FILE_MULTIPLICATIVE = resource_filename(
    "pyproteininference", "../tests/data/test_params_multiplicative_custom_score.yaml"
)

TARGET_FILE_ADDITIVE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target_additive.txt")
DECOY_FILE_ADDITIVE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy_additive.txt")
PARAMETER_FILE_ADDITIVE = resource_filename(
    "pyproteininference", "../tests/data/test_params_additive_custom_score.yaml"
)


class TestCustomRestrict(TestCase):
    def test_override_params_and_mult_restrict(self):

        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE_MULTIPLICATIVE)

        # Redefine some of the param options...
        protein_inference_parameters.restrict_pep = 0.9
        protein_inference_parameters.restrict_q = 0.1
        protein_inference_parameters.restrict_custom = 0.0009

        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE_MULTIPLICATIVE,
            decoy_file=DECOY_FILE_MULTIPLICATIVE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        data.restrict_psm_data()

        # Make sure we restrict by custom score
        self.assertEqual(len(data.main_data_restricted), 11)

        # Make sure other restrict params get set to None
        self.assertIsNone(protein_inference_parameters.restrict_pep)
        self.assertIsNone(protein_inference_parameters.restrict_q)

    def test_override_params_and_add_restrict(self):

        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE_ADDITIVE)

        # Redefine some of the param options...
        protein_inference_parameters.restrict_pep = 0.9
        protein_inference_parameters.restrict_q = 0.1
        protein_inference_parameters.restrict_custom = 6

        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE_ADDITIVE,
            decoy_file=DECOY_FILE_ADDITIVE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        data.restrict_psm_data()

        # Make sure custom score gets restricted
        self.assertEqual(len(data.main_data_restricted), 9)

        # Make sure parameters have been set to None
        self.assertIsNone(protein_inference_parameters.restrict_pep)
        self.assertIsNone(protein_inference_parameters.restrict_q)



import collections
import logging
import sys

from pyproteininference.inference import Inference
from pyproteininference.physical import Protein, Psm
from pyproteininference.scoring import Score

logger = logging.getLogger(__name__)

# set up our logger
logging.basicConfig(
    stream=sys.stderr,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class DataStore(object):
    """
    The following Class serves as the data storage object for a protein inference analysis
    The class serves as a central point that is accessed at virtually every PI processing step


    Attributes:
        main_data_form (list): List of unrestricted Psm objects.
        parameter_file_object (ProteinInferenceParameter): protein inference parameter
            [object][pyproteininference.parameters.ProteinInferenceParameter].
        restricted_peptides (list): List of non flaking peptide strings present in the current analysis.
        main_data_restricted (list): List of restricted [Psm][pyproteininference.physical.Psm] objects.
            Restriction is based on the parameter_file_object and the object is created by function
                [restrict_psm_data][pyproteininference.datastore.DataStore.restrict_psm_data].
        scored_proteins (list): List of scored [Protein][pyproteininference.physical.Protein] objects.
            Output from scoring methods from [scoring][pyproteininference.scoring].
        grouped_scored_proteins (list): List of scored [Protein][pyproteininference.physical.Protein]
            objects that have been grouped and sorted. Output from
                [run_inference][pyproteininference.inference.Inference.run_inference] method.
        scoring_input (list): List of non-scored [Protein][pyproteininference.physical.Protein] objects.
            Output from [create_scoring_input][pyproteininference.datastore.DataStore.create_scoring_input].
        picked_proteins_scored (list): List of [Protein][pyproteininference.physical.Protein] objects that pass
            the protein picker algorithm ([protein_picker][pyproteininference.datastore.DataStore.protein_picker]).
        picked_proteins_removed (list): List of [Protein][pyproteininference.physical.Protein] objects that do not
            pass the protein picker algorithm ([protein_picker][pyproteininference.datastore.DataStore.protein_picker]).
        protein_peptide_dictionary (collections.defaultdict): Dictionary of protein strings (keys) that map to sets
            of peptide strings based on the peptides and proteins found in the search. Protein -> set(Peptides).
        peptide_protein_dictionary (collections.defaultdict): Dictionary of peptide strings (keys) that map to sets
            of protein strings based on the peptides and proteins found in the search. Peptide -> set(Proteins).
        high_low_better (str): Variable that indicates whether a higher or a lower protein score is better.
            This is necessary to sort Protein objects by score properly. Can either be "higher" or "lower".
        psm_score (str): Variable that indicates the [Psm][pyproteininference.physical.Psm]
            score being used in the analysis to generate [Protein][pyproteininference.physical.Protein] scores.
        protein_score (str): String to indicate the protein score method used.
        short_protein_score (str): Short String to indicate the protein score method used.
        protein_group_objects (list): List of scored [ProteinGroup][pyproteininference.physical.ProteinGroup]
            objects that have been grouped and sorted. Output from
             [run_inference][pyproteininference.inference.Inference.run_inference] method.
        decoy_symbol (str): String that is used to differentiate between decoy proteins and target proteins. Ex: "##".
        digest (Digest): [Digest object][pyproteininference.in_silico_digest.Digest].
        SCORE_MAPPER (dict): Dictionary that maps potential scores in input files to internal score names.
        CUSTOM_SCORE_KEY (str): String that indicates a custom score is being used.

    """

    SCORE_MAPPER = {
        "q_value": "qvalue",
        "pep_value": "pepvalue",
        "perc_score": "percscore",
        "score": "percscore",
        "q-value": "qvalue",
        "posterior_error_prob": "pepvalue",
        "posterior_error_probability": "pepvalue",
    }

    CUSTOM_SCORE_KEY = "custom_score"

    HIGHER_PSM_SCORE = "higher"
    LOWER_PSM_SCORE = "lower"

    def __init__(self, reader, digest, validate=True):
        """

        Args:
            reader (Reader): Reader object [Reader][pyproteininference.reader.Reader].
            digest (Digest): Digest object
                [Digest][pyproteininference.in_silico_digest.Digest].
            validate (bool): True/False to indicate if the input data should be validated.

        Example:
            >>> pyproteininference.datastore.DataStore(reader = reader, digest=digest)


        """
        # If the reader class is from a percolator.psms then define main_data_form as reader.psms
        # main_data_form is the starting point for all other analyses
        self._init_validate(reader=reader)

        self.parameter_file_object = reader.parameter_file_object  # Parameter object
        self.main_data_restricted = None  # PSM data post restriction
        self.scored_proteins = []  # List of scored Protein objects
        self.grouped_scored_proteins = []  # List of sorted scored Protein objects
        self.scoring_input = None  # List of non scored Protein objects
        self.picked_proteins_scored = None  # List of Protein objects after picker algorithm
        self.picked_proteins_removed = None  # Protein objects removed via picker
        self.protein_peptide_dictionary = None
        self.peptide_protein_dictionary = None
        self.high_low_better = None  # Variable that indicates whether a higher or lower protein score is better
        self.psm_score = None  # PSM Score used
        self.protein_score = None
        self.short_protein_score = None
        self.protein_group_objects = []  # List of sorted protein group objects
        self.decoy_symbol = self.parameter_file_object.decoy_symbol  # Decoy symbol from parameter file
        self.digest = digest  # Digest object

        # Run Checks and Validations
        if validate:
            self.validate_psm_data()
            self.validate_digest()
            self.check_data_consistency()

        # Run method to fix our parameter object if necessary
        self.parameter_file_object.fix_parameters_from_datastore(data=self)

    def get_sorted_identifiers(self, scored=True):
        """
        Retrieves a sorted list of protein strings present in the analysis.

        Args:
            scored (bool): True/False to indicate if we should return scored or non-scored identifiers.

        Returns:
            list: List of sorted protein identifier strings.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> sorted_proteins = data.get_sorted_identifiers(scored=True)
        """

        if scored:
            self._validate_scored_proteins()
            if self.picked_proteins_scored:
                proteins = set([x.identifier for x in self.picked_proteins_scored])
            else:
                proteins = set([x.identifier for x in self.scored_proteins])
        else:
            self._validate_scoring_input()
            proteins = [x.identifier for x in self.scoring_input]

        all_sp_proteins = set(self.digest.swiss_prot_protein_set)

        our_target_sp_proteins = sorted([x for x in proteins if x in all_sp_proteins and self.decoy_symbol not in x])
        our_decoy_sp_proteins = sorted([x for x in proteins if x in all_sp_proteins and self.decoy_symbol in x])

        our_target_tr_proteins = sorted(
            [x for x in proteins if x not in all_sp_proteins and self.decoy_symbol not in x]
        )
        our_decoy_tr_proteins = sorted([x for x in proteins if x not in all_sp_proteins and self.decoy_symbol in x])

        our_proteins_sorted = (
            our_target_sp_proteins + our_decoy_sp_proteins + our_target_tr_proteins + our_decoy_tr_proteins
        )

        return our_proteins_sorted

    @classmethod
    def sort_protein_group_objects(cls, protein_group_objects, higher_or_lower):
        """
        Class Method to sort a list of [ProteinGroup][pyproteininference.physical.ProteinGroup] objects by
        score and number of peptides.

        Args:
            protein_group_objects (list): list of [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.
            higher_or_lower (str): String to indicate if a "higher" or "lower" protein score is "better".

        Returns:
            list: list of sorted [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        Example:
            >>> list_of_group_objects = pyproteininference.datastore.DataStore.sort_protein_group_objects(
            >>>     protein_group_objects=list_of_group_objects, higher_or_lower="higher"
            >>> )
        """
        if higher_or_lower == cls.LOWER_PSM_SCORE:

            protein_group_objects = sorted(
                protein_group_objects,
                key=lambda k: (
                    k.proteins[0].score,
                    -k.proteins[0].num_peptides,
                ),
                reverse=False,
            )
        elif higher_or_lower == cls.HIGHER_PSM_SCORE:

            protein_group_objects = sorted(
                protein_group_objects,
                key=lambda k: (
                    k.proteins[0].score,
                    k.proteins[0].num_peptides,
                ),
                reverse=True,
            )

        return protein_group_objects

    @classmethod
    def sort_protein_objects(cls, grouped_protein_objects, higher_or_lower):
        """
        Class Method to sort a list of [Protein][pyproteininference.physical.Protein] objects by score and number of
        peptides.

        Args:
            grouped_protein_objects (list): list of [Protein][pyproteininference.physical.Protein] objects.
            higher_or_lower (str): String to indicate if a "higher" or "lower" protein score is "better".

        Returns:
            list: list of sorted [Protein][pyproteininference.physical.Protein] objects.

        Example:
            >>> scores_grouped = pyproteininference.datastore.DataStore.sort_protein_objects(
            >>>     grouped_protein_objects=scores_grouped, higher_or_lower="higher"
            >>> )
        """
        if higher_or_lower == cls.LOWER_PSM_SCORE:
            grouped_protein_objects = sorted(
                grouped_protein_objects,
                key=lambda k: (k[0].score, -k[0].num_peptides),
                reverse=False,
            )
        if higher_or_lower == cls.HIGHER_PSM_SCORE:
            grouped_protein_objects = sorted(
                grouped_protein_objects,
                key=lambda k: (k[0].score, k[0].num_peptides),
                reverse=True,
            )
        return grouped_protein_objects

    @classmethod
    def sort_protein_sub_groups(cls, protein_list, higher_or_lower):
        """
        Method to sort protein sub lists.

        Args:
            protein_list (list): List of [Protein][pyproteininference.physical.Protein] objects to be sorted.
            higher_or_lower (str): String to indicate if a "higher" or "lower" protein score is "better".

        Returns:
            list: List of [Protein][pyproteininference.physical.Protein] objects to be sorted by score and number of
            peptides.

        """

        # Sort the groups based on higher or lower indication, secondarily sort the groups based on number of unique
        # peptides
        # We use the index [1:] as we do not wish to sort the lead protein...
        if higher_or_lower == cls.LOWER_PSM_SCORE:
            protein_list[1:] = sorted(
                protein_list[1:],
                key=lambda k: (float(k.score), -float(k.num_peptides)),
                reverse=False,
            )
        if higher_or_lower == cls.HIGHER_PSM_SCORE:
            protein_list[1:] = sorted(
                protein_list[1:],
                key=lambda k: (float(k.score), float(k.num_peptides)),
                reverse=True,
            )

        return protein_list

    def get_psm_data(self):
        """
        Method to retrieve a list of [Psm][pyproteininference.physical.Psm] objects.
        Retrieves restricted data if the data has been restricted or all of the data if the data has
        not been restricted.

        Returns:
            list: list of [Psm][pyproteininference.physical.Psm] objects.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> psm_data = data.get_psm_data()
        """
        if not self.main_data_restricted and not self.main_data_form:
            raise ValueError(
                "Both main_data_restricted and main_data_form variables are empty. Please re-load the DataStore "
                "object with a properly loaded Reader object."
            )

        if self.main_data_restricted:
            psm_data = self.main_data_restricted
        else:
            psm_data = self.main_data_form

        return psm_data

    def get_protein_data(self):
        """
        Method to retrieve a list of [Protein][pyproteininference.physical.Protein] objects.
        Retrieves picked and scored data if the data has been picked and scored or just the scored data if the data has
         not been picked.

        Returns:
            list: list of [Protein][pyproteininference.physical.Protein] objects.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> # Data must ben ran through a pyproteininference.scoring.Score method
            >>> protein_data = data.get_protein_data()
        """

        if self.picked_proteins_scored:
            scored_proteins = self.picked_proteins_scored
        else:
            scored_proteins = self.scored_proteins

        return scored_proteins

    def get_protein_identifiers_from_psm_data(self):
        """
        Method to retrieve a list of lists of all possible protein identifiers from the psm data.

        Returns:
            list: list of lists of protein strings.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> protein_strings = data.get_protein_identifiers_from_psm_data()
        """
        psm_data = self.get_psm_data()

        proteins = [x.possible_proteins for x in psm_data]

        return proteins

    def get_q_values(self):
        """
        Method to retrieve a list of all q values for all PSMs.

        Returns:
            list: list of floats (q values).

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> q = data.get_q_values()
        """
        psm_data = self.get_psm_data()

        q_values = [x.qvalue for x in psm_data]

        return q_values

    def get_pep_values(self):
        """
        Method to retrieve a list of all posterior error probabilities for all PSMs.

        Returns:
            list: list of floats (pep values).

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> pep = data.get_pep_values()
        """
        psm_data = self.get_psm_data()

        pep_values = [x.pepvalue for x in psm_data]

        return pep_values

    def get_protein_information_dictionary(self):
        """
        Method to retrieve a dictionary of scores for each peptide.

        Returns:
            dict: dictionary of scores for each protein.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> protein_dict = data.get_protein_information_dictionary()
        """
        psm_data = self.get_psm_data()

        protein_psm_score_dictionary = collections.defaultdict(list)

        # Loop through all Psms
        for psms in psm_data:
            # Loop through all proteins
            for prots in psms.possible_proteins:
                protein_psm_score_dictionary[prots].append(
                    {
                        "peptide": psms.identifier,
                        "Qvalue": psms.qvalue,
                        "PosteriorErrorProbability": psms.pepvalue,
                        "Percscore": psms.percscore,
                    }
                )

        return protein_psm_score_dictionary

    def restrict_psm_data(self, remove1pep=True):
        """
        Method to restrict the input of [Psm][pyproteininference.physical.Psm]  objects.
        This method is central to the pyproteininference module and is able to restrict the Psm data by:
        Q value, Pep Value, Percolator Score, Peptide Length, and Custom Score Input.
        Restriction values are pulled from
        the [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter]
        object.

        This method sets the `main_data_restricted` and `restricted_peptides` Attributes for the DataStore object.

        Args:
            remove1pep (bool): True/False on whether or not to remove PEP values that equal 1 even if other restrictions
                are set to not restrict.

        Returns:
            None:

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> data.restrict_psm_data(remove1pep=True)
        """

        # Validate that we have the main data variable
        self._validate_main_data_form()

        logger.info("Restricting PSM data")

        peptide_length = self.parameter_file_object.restrict_peptide_length
        posterior_error_prob_threshold = self.parameter_file_object.restrict_pep
        q_value_threshold = self.parameter_file_object.restrict_q
        custom_threshold = self.parameter_file_object.restrict_custom

        main_psm_data = self.main_data_form
        logger.info("Length of main data: {}".format(len(self.main_data_form)))
        # If restrict_main_data is called, we automatically discard everything that has a PEP of 1
        if remove1pep and posterior_error_prob_threshold:
            main_psm_data = [x for x in main_psm_data if x.pepvalue != 1]

        # Restrict peptide length and posterior error probability
        if peptide_length and posterior_error_prob_threshold and not q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if len(psms.stripped_peptide) >= peptide_length and psms.pepvalue < float(
                    posterior_error_prob_threshold
                ):
                    restricted_data.append(psms)

        # Restrict peptide length only
        if peptide_length and not posterior_error_prob_threshold and not q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if len(psms.stripped_peptide) >= peptide_length:
                    restricted_data.append(psms)

        # Restrict peptide length, posterior error probability, and qvalue
        if peptide_length and posterior_error_prob_threshold and q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if (
                    len(psms.stripped_peptide) >= peptide_length
                    and psms.pepvalue < float(posterior_error_prob_threshold)
                    and psms.qvalue < float(q_value_threshold)
                ):
                    restricted_data.append(psms)

        # Restrict peptide length and qvalue
        if peptide_length and not posterior_error_prob_threshold and q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if len(psms.stripped_peptide) >= peptide_length and psms.qvalue < float(q_value_threshold):
                    restricted_data.append(psms)

        # Restrict posterior error probability and q value
        if not peptide_length and posterior_error_prob_threshold and q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if psms.pepvalue < float(posterior_error_prob_threshold) and psms.qvalue < float(q_value_threshold):
                    restricted_data.append(psms)

        # Restrict qvalue only
        if not peptide_length and not posterior_error_prob_threshold and q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if psms.qvalue < float(q_value_threshold):
                    restricted_data.append(psms)

        # Restrict posterior error probability only
        if not peptide_length and posterior_error_prob_threshold and not q_value_threshold:
            restricted_data = []
            for psms in main_psm_data:
                if psms.pepvalue < float(posterior_error_prob_threshold):
                    restricted_data.append(psms)

        # Restrict nothing... (only PEP gets restricted - takes everything less than 1)
        if not peptide_length and not posterior_error_prob_threshold and not q_value_threshold:
            restricted_data = main_psm_data

        if custom_threshold:
            custom_restricted = []
            if self.parameter_file_object.psm_score_type == Score.MULTIPLICATIVE_SCORE_TYPE:
                for psms in restricted_data:
                    if psms.custom_score <= custom_threshold:
                        custom_restricted.append(psms)

            if self.parameter_file_object.psm_score_type == Score.ADDITIVE_SCORE_TYPE:
                for psms in restricted_data:
                    if psms.custom_score >= custom_threshold:
                        custom_restricted.append(psms)

            restricted_data = custom_restricted

        self.main_data_restricted = restricted_data

        logger.info("Length of restricted data: {}".format(len(restricted_data)))

        self.restricted_peptides = [x.non_flanking_peptide for x in restricted_data]

    def create_scoring_input(self):
        """
        Method to create the scoring input.
        This method initializes a list of [Protein][pyproteininference.physical.Protein] objects to get them ready
        to be scored by [Score][pyproteininference.scoring.Score] methods.
        This method also takes into account the inference type and aggregates peptides -> proteins accordingly.

        This method sets the `scoring_input` and `score` Attributes for the DataStore object.

        The score selected comes from the protein inference parameter object.

        Returns:
            None:

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> data.create_scoring_input()
        """

        logger.info("Creating Scoring Input")

        psm_data = self.get_psm_data()

        protein_psm_dict = collections.defaultdict(list)

        try:
            score_key = self.SCORE_MAPPER[self.parameter_file_object.psm_score]
        except KeyError:
            score_key = self.CUSTOM_SCORE_KEY

        if self.parameter_file_object.inference_type != Inference.PEPTIDE_CENTRIC:
            # Loop through all Psms
            for psms in psm_data:
                psms.assign_main_score(score=score_key)
                # Loop through all proteins
                for prots in psms.possible_proteins:
                    protein_psm_dict[prots].append(psms)

        else:
            self.peptide_to_protein_dictionary()
            sp_proteins = self.digest.swiss_prot_protein_set
            for psms in psm_data:

                # Assign main score
                psms.assign_main_score(score=score_key)
                protein_set = self.peptide_protein_dictionary[psms.non_flanking_peptide]
                # Sort protein_set by sp-alpha, decoy-sp-alpha, tr-alpha, decoy-tr-alpha
                sorted_protein_list = self.sort_protein_strings(
                    protein_string_list=protein_set,
                    sp_proteins=sp_proteins,
                    decoy_symbol=self.parameter_file_object.decoy_symbol,
                )
                # Restrict the number of identifiers by the value in param file max_identifiers_peptide_centric
                sorted_protein_list = sorted_protein_list[: self.parameter_file_object.max_identifiers_peptide_centric]
                protein_name = ";".join(sorted_protein_list)
                protein_psm_dict[protein_name].append(psms)

        protein_list = []
        for pkey in sorted(protein_psm_dict.keys()):
            protein_object = Protein(identifier=pkey)
            protein_object.psms = protein_psm_dict[pkey]
            protein_object.raw_peptides = set([x.identifier for x in protein_psm_dict[pkey]])
            protein_list.append(protein_object)

        self.psm_score = self.parameter_file_object.psm_score
        self.scoring_input = protein_list

    def protein_to_peptide_dictionary(self):
        """
        Method that returns a map of protein strings to sets of peptide strings and is essentially half
         of a BiPartite graph.
        This method sets the `protein_peptide_dictionary` Attribute for the DataStore object.

        Returns:
            collections.defaultdict: Dictionary of protein strings (keys) that map to sets of peptide strings based
            on the peptides and proteins found in the search. Protein -> set(Peptides).

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> protein_peptide_dict = data.protein_to_peptide_dictionary()
        """
        psm_data = self.get_psm_data()

        res_pep_set = set(self.restricted_peptides)
        default_dict_proteins = collections.defaultdict(set)
        for peptide_objects in psm_data:
            for prots in peptide_objects.possible_proteins:
                cur_peptide = peptide_objects.non_flanking_peptide
                if cur_peptide in res_pep_set:
                    default_dict_proteins[prots].add(cur_peptide)

        self.protein_peptide_dictionary = default_dict_proteins

        return default_dict_proteins

    def peptide_to_protein_dictionary(self):
        """
        Method that returns a map of peptide strings to sets of protein strings and is essentially half of a
        BiPartite graph.
        This method sets the `peptide_protein_dictionary` Attribute for the DataStore object.

        Returns:
            collections.defaultdict: Dictionary of peptide strings (keys) that map to sets of protein strings based
                on the peptides and proteins found in the search. Peptide -> set(Proteins).

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> peptide_protein_dict = data.peptide_to_protein_dictionary()
        """
        psm_data = self.get_psm_data()

        res_pep_set = set(self.restricted_peptides)
        default_dict_peptides = collections.defaultdict(set)
        for peptide_objects in psm_data:
            for prots in peptide_objects.possible_proteins:
                cur_peptide = peptide_objects.non_flanking_peptide
                if cur_peptide in res_pep_set:
                    default_dict_peptides[cur_peptide].add(prots)
                else:
                    pass

        self.peptide_protein_dictionary = default_dict_peptides

        return default_dict_peptides

    def unique_to_leads_peptides(self):
        """
        Method to retrieve peptides that are unique based on the data from the searches
        (Not based on the database digestion).

        Returns:
            set: a Set of peptide strings

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> unique_peps = data.unique_to_leads_peptides()
        """
        if self.grouped_scored_proteins:
            lead_peptides = [list(x[0].peptides) for x in self.grouped_scored_proteins]
            flat_peptides = [item for sublist in lead_peptides for item in sublist]
            counted_peps = collections.Counter(flat_peptides)
            unique_to_leads_peptides = set([x for x in counted_peps if counted_peps[x] == 1])
        else:
            unique_to_leads_peptides = set()

        return unique_to_leads_peptides

    def higher_or_lower(self):
        """
        Method to determine if a higher or lower score is better for a given combination of score input and score type.

        This method sets the `high_low_better` Attribute for the DataStore object.

        This method depends on the output from the Score class to be sorted properly from best to worst score.

        Returns:
            str: String indicating "higher" or "lower" depending on if a higher or lower score is a
                better protein score.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> high_low = data.higher_or_lower()
        """

        if not self.high_low_better:
            logger.info("Determining If a higher or lower score is better based on scored proteins")
            worst_score = self.scored_proteins[-1].score
            best_score = self.scored_proteins[0].score

            if float(best_score) > float(worst_score):
                higher_or_lower = self.HIGHER_PSM_SCORE

            if float(best_score) < float(worst_score):
                higher_or_lower = self.LOWER_PSM_SCORE

            logger.info("best score = {}".format(best_score))
            logger.info("worst score = {}".format(worst_score))

            if best_score == worst_score:
                raise ValueError(
                    "Best and Worst scores were identical, equal to {}. Score type {} produced the error, "
                    "please change psm_score type.".format(best_score, self.psm_score)
                )

            self.high_low_better = higher_or_lower

        else:
            higher_or_lower = self.high_low_better

        return higher_or_lower

    def get_protein_identifiers(self, data_form):
        """
        Method to retrieve the protein string identifiers.

        Args:
            data_form (str): Can be one of the following: "main", "restricted", "picked", "picked_removed".

        Returns:
            list: list of protein identifier strings.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> protein_strings = data.get_protein_identifiers(data_form="main")
        """
        if data_form == "main":
            # All the data (unrestricted)
            data_to_select = self.main_data_form
            prots = [[x.possible_proteins] for x in data_to_select]
            proteins = prots

        if data_form == "restricted":
            # Proteins that pass certain restriction criteria (peptide length, pep, qvalue)
            data_to_select = self.main_data_restricted
            prots = [[x.possible_proteins] for x in data_to_select]
            proteins = prots

        if data_form == "picked":
            # Here we look at proteins that are 'picked' (aka the proteins that beat out their matching target/decoy)
            data_to_select = self.picked_proteins_scored
            prots = [x.identifier for x in data_to_select]
            proteins = prots

        if data_form == "picked_removed":
            # Here we look at the proteins that were removed due to picking (aka the proteins that
            # have a worse score than their target/decoy counterpart)
            data_to_select = self.picked_proteins_removed
            prots = [x.identifier for x in data_to_select]
            proteins = prots

        return proteins

    def get_protein_information(self, protein_string):
        """
        Method to retrieve attributes for a specific scored protein.

        Args:
            protein_string (str): Protein Identifier String.

        Returns:
            list: list of protein attributes.

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> protein_attr = data.get_protein_information(protein_string="RAF1_HUMAN|P04049")
        """
        all_scored_protein_data = self.scored_proteins
        identifiers = [x.identifier for x in all_scored_protein_data]
        protein_scores = [x.score for x in all_scored_protein_data]
        groups = [x.group_identification for x in all_scored_protein_data]
        reviewed = [x.reviewed for x in all_scored_protein_data]
        peptides = [x.peptides for x in all_scored_protein_data]
        # Peptide scores currently broken...
        peptide_scores = [x.peptide_scores for x in all_scored_protein_data]
        picked = [x.picked for x in all_scored_protein_data]
        num_peptides = [x.num_peptides for x in all_scored_protein_data]

        main_index = identifiers.index(protein_string)

        list_structure = [
            [
                "identifier",
                "protein_score",
                "groups",
                "reviewed",
                "peptides",
                "peptide_scores",
                "picked",
                "num_peptides",
            ]
        ]
        list_structure.append([protein_string])
        list_structure[-1].append(protein_scores[main_index])
        list_structure[-1].append(groups[main_index])
        list_structure[-1].append(reviewed[main_index])
        list_structure[-1].append(peptides[main_index])
        list_structure[-1].append(peptide_scores[main_index])
        list_structure[-1].append(picked[main_index])
        list_structure[-1].append(num_peptides[main_index])

        return list_structure

    def exclude_non_distinguishing_peptides(self, protein_subset_type="hard"):
        """
        Method to Exclude peptides that are not distinguishing on either the search or database level.

        The method sets the `scoring_input` and `restricted_peptides` variables for the DataStore object.

        Args:
            protein_subset_type (str): Either "hard" or "soft". Hard will select distinguishing peptides based on
                the database digestion. "soft" will only use peptides identified in the search.

        Returns:
            None:

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> data.exclude_non_distinguishing_peptides(protein_subset_type="hard")
        """

        logger.info("Applying Exclusion Model")

        our_proteins_sorted = self.get_sorted_identifiers(scored=False)

        if protein_subset_type == "hard":
            # Hard protein subsetting defines protein subsets on the digest level (Entire protein is used)
            # This is how Percolator PI does subsetting
            peptides = [self.digest.protein_to_peptide_dictionary[x] for x in our_proteins_sorted]
        elif protein_subset_type == "soft":
            # Soft protein subsetting defines protein subsets on the Peptides identified from the search
            peptides = [set(x.raw_peptides) for x in self.scoring_input]
        else:
            # If neither is dfined we do "hard" exclusion
            peptides = [self.digest.protein_to_peptide_dictionary[x] for x in our_proteins_sorted]

        # Get frozen set of peptides....
        # We will also have a corresponding list of proteins...
        # They will have the same index...
        peptide_sets = [frozenset(e) for e in peptides]
        # Find a way to sort this list of sets...
        # We can sort the sets if we sort proteins from above...
        logger.info("{} number of peptide sets".format(len(peptide_sets)))
        non_subset_peptide_sets = set()
        i = 0
        # Get all peptide sets that are not a subset...
        while peptide_sets:
            i = i + 1
            peptide_set = peptide_sets.pop()
            if any(peptide_set.issubset(s) for s in peptide_sets) or any(
                peptide_set.issubset(s) for s in non_subset_peptide_sets
            ):
                continue
            else:
                non_subset_peptide_sets.add(peptide_set)
            if i % 10000 == 0:
                logger.info("Parsed {} Peptide Sets".format(i))

        logger.info("Parsed {} Peptide Sets".format(i))

        # Get their index from peptides which is the initial list of sets...
        list_of_indeces = []
        for pep_sets in non_subset_peptide_sets:
            ind = peptides.index(pep_sets)
            list_of_indeces.append(ind)

        non_subset_proteins = set([our_proteins_sorted[x] for x in list_of_indeces])

        logger.info("Removing direct subset Proteins from the data")
        # Remove all proteins from scoring input that are a subset of another protein...
        self.scoring_input = [x for x in self.scoring_input if x.identifier in non_subset_proteins]

        logger.info("{} proteins in scoring input after removing subset proteins".format(len(self.scoring_input)))

        # For all the proteins that are not a complete subset of another protein...
        # Get the raw peptides...
        raw_peps = [x.raw_peptides for x in self.scoring_input if x.identifier in non_subset_proteins]

        # Make the raw peptides a flat list
        flat_peptides = [Psm.split_peptide(peptide_string=item) for sublist in raw_peps for item in sublist]

        # Count the number of peptides in this list...
        # This is the number of proteins this peptide maps to....
        counted_peptides = collections.Counter(flat_peptides)

        # If the count is greater than 1... exclude the protein entirely from scoring input... :)
        raw_peps_good = set([x for x in counted_peptides.keys() if counted_peptides[x] <= 1])

        # Alter self.scoring_input by removing psms and peptides that are not in raw_peps_good
        current_score_input = list(self.scoring_input)
        for j in range(len(current_score_input)):
            k = j + 1
            psm_list = []
            new_raw_peptides = []
            current_psms = current_score_input[j].psms
            current_raw_peptides = current_score_input[j].raw_peptides

            for psm_scores in current_psms:
                if psm_scores.non_flanking_peptide in raw_peps_good:
                    psm_list.append(psm_scores)

            for rp in current_raw_peptides:
                if Psm.split_peptide(peptide_string=rp) in raw_peps_good:
                    new_raw_peptides.append(rp)

            current_score_input[j].psms = psm_list
            current_score_input[j].raw_peptides = new_raw_peptides

            if k % 10000 == 0:
                logger.info("Redefined {} Peptide Sets".format(k))

        logger.info("Redefined {} Peptide Sets".format(j))

        filtered_score_input = [x for x in current_score_input if x.psms]

        self.scoring_input = filtered_score_input

        # Recompute the flat peptides
        raw_peps = [x.raw_peptides for x in self.scoring_input if x.identifier in non_subset_proteins]

        # Make the raw peptides a flat list
        new_flat_peptides = set([Psm.split_peptide(peptide_string=item) for sublist in raw_peps for item in sublist])

        self.scoring_input = [x for x in self.scoring_input if x.psms]

        self.restricted_peptides = [x for x in self.restricted_peptides if x in new_flat_peptides]

    def protein_picker(self):
        """
        Method to run the protein picker algorithm.

        Proteins must be scored first with [score_psms][pyproteininference.scoring.Score.score_psms].

        The algorithm will match target and decoy proteins identified from the PSMs from the search.
        If a target and matching decoy is found then target/decoy competition is performed.
        In the Target/Decoy pair the protein with the better score is kept and the one with the worse score is
        discarded from the analysis.

        The method sets the `picked_proteins_scored` and `picked_proteins_removed` variables for
        the DataStore object.

        Returns:
            None:

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> data.protein_picker()
        """

        self._validate_scored_proteins()

        logger.info("Running Protein Picker")

        # Use higher or lower class to determine if a higher protein score or lower protein score is better
        # based on the scoring method used
        higher_or_lower = self.higher_or_lower()
        # Here we determine if a lower or higher score is better
        # Since all input is ordered from best to worst we can do the following

        index_to_remove = []
        # data.scored_proteins is simply a list of Protein objects...
        # Create list of all decoy proteins
        decoy_proteins = [x.identifier for x in self.scored_proteins if self.decoy_symbol in x.identifier]
        # Create a list of all potential matching targets (some of these may not exist in the search)
        matching_targets = [x.replace(self.decoy_symbol, "") for x in decoy_proteins]

        # Create a list of all the proteins from the scored data
        all_proteins = [x.identifier for x in self.scored_proteins]
        logger.info("{} proteins scored".format(len(all_proteins)))

        total_targets = []
        total_decoys = []
        decoys_removed = []
        targets_removed = []
        # Loop over all decoys identified in the search
        logger.info("Picking Proteins...")
        for i in range(len(decoy_proteins)):
            cur_decoy_index = all_proteins.index(decoy_proteins[i])
            cur_decoy_protein_object = self.scored_proteins[cur_decoy_index]
            total_decoys.append(cur_decoy_protein_object.identifier)

            # Try, Except here because the matching target to the decoy may not be a result from the search
            try:
                cur_target_index = all_proteins.index(matching_targets[i])
                cur_target_protein_object = self.scored_proteins[cur_target_index]
                total_targets.append(cur_target_protein_object.identifier)

                if higher_or_lower == self.HIGHER_PSM_SCORE:
                    if cur_target_protein_object.score > cur_decoy_protein_object.score:
                        index_to_remove.append(cur_decoy_index)
                        decoys_removed.append(cur_decoy_index)
                        cur_target_protein_object.picked = True
                        cur_decoy_protein_object.picked = False
                    else:
                        index_to_remove.append(cur_target_index)
                        targets_removed.append(cur_target_index)
                        cur_decoy_protein_object.picked = True
                        cur_target_protein_object.picked = False

                if higher_or_lower == self.LOWER_PSM_SCORE:
                    if cur_target_protein_object.score < cur_decoy_protein_object.score:
                        index_to_remove.append(cur_decoy_index)
                        decoys_removed.append(cur_decoy_index)
                        cur_target_protein_object.picked = True
                        cur_decoy_protein_object.picked = False
                    else:
                        index_to_remove.append(cur_target_index)
                        targets_removed.append(cur_target_index)
                        cur_decoy_protein_object.picked = True
                        cur_target_protein_object.picked = False
            except ValueError:
                pass

        logger.info("{} total decoy proteins".format(len(total_decoys)))
        logger.info("{} matching target proteins also found in search".format(len(total_targets)))
        logger.info("{} decoy proteins to be removed".format(len(decoys_removed)))
        logger.info("{} target proteins to be removed".format(len(targets_removed)))

        logger.info("Removing Lower Scoring Proteins...")
        picked_list = []
        removed_proteins = []
        for protein_objects in self.scored_proteins:
            if protein_objects.picked:
                picked_list.append(protein_objects)
            else:
                removed_proteins.append(protein_objects)
        self.picked_proteins_scored = picked_list
        self.picked_proteins_removed = removed_proteins
        logger.info("Finished Removing Proteins")

    def calculate_q_values(self, regular=True):
        """
        Method calculates Q values FDR on the lead protein in the group on the `protein_group_objects`
        instance variable.
        FDR is calculated As (2*decoys)/total if regular is set to True and is
        (decoys)/total if regular is set to False.

        This method updates the `protein_group_objects` for the DataStore object by updating
        the q_value variable of the [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        Returns:
            None:

        Example:
            >>> data = pyproteininference.datastore.DataStore(reader = reader, digest=digest)
            >>> # Data must be scored first
            >>> data.calculate_q_values()
        """

        self._validate_protein_group_objects()

        logger.info("Calculating Q values from the protein group objects")

        # pick out the lead scoring protein for each group... lead score is at 0 position
        lead_score = [x.proteins[0] for x in self.protein_group_objects]
        # Now pick out only the lead protein identifiers
        lead_proteins = [x.identifier for x in lead_score]

        lead_proteins.reverse()

        logger.info("Calculating FDRs")
        fdr_list = []
        for i in range(len(lead_proteins)):
            binary_decoy_target_list = [1 if self.decoy_symbol in elem else 0 for elem in lead_proteins]
            total = len(lead_proteins)
            decoys = sum(binary_decoy_target_list)
            # Calculate FDR at every step starting with the entire list...
            # Delete first entry (worst score) every time we go through a cycle
            if regular:
                fdr = (2 * decoys) / (float(total))
            else:
                fdr = (decoys) / (float(total))
            fdr_list.append(fdr)
            del lead_proteins[0]

        qvalue_list = []
        new_fdr_list = []
        logger.info("Calculating Q Values")
        for fdrs in fdr_list:
            new_fdr_list.append(fdrs)
            qvalue = min(new_fdr_list)
            # qvalue = fdrs
            qvalue_list.append(qvalue)

        qvalue_list.reverse()

        logger.info("Assigning Q Values")
        for k in range(len(self.protein_group_objects)):
            self.protein_group_objects[k].q_value = qvalue_list[k]

        fdr_restricted = [x for x in self.protein_group_objects if x.q_value <= self.parameter_file_object.fdr]

        fdr_restricted_set = [self.grouped_scored_proteins[x] for x in range(len(fdr_restricted))]

        onehitwonders = []
        for groups in fdr_restricted_set:
            if int(groups[0].num_peptides) == 1:
                onehitwonders.append(groups[0])

        logger.info(
            "Protein Group leads that pass with more than 1 PSM with a {} FDR = {}".format(
                self.parameter_file_object.fdr,
                str(len(fdr_restricted_set) - len(onehitwonders)),
            )
        )
        logger.info(
            "Protein Group lead One hit Wonders that pass {} FDR = {}".format(
                self.parameter_file_object.fdr, len(onehitwonders)
            )
        )

        logger.info(
            "Number of Protein groups that pass a {} percent FDR: {}".format(
                str(self.parameter_file_object.fdr * 100), len(fdr_restricted_set)
            )
        )

        logger.info("Finished Q value Calculation")

    def validate_psm_data(self):
        """
        Method that validates the PSM data.
        """
        self._validate_decoys_from_data()
        self._validate_isoform_from_data()

    def validate_digest(self):
        """
        Method that validates the [Digest object][pyproteininference.in_silico_digest.Digest].
        """
        self._validate_reviewed_v_unreviewed()
        self._check_target_decoy_split()

    def check_data_consistency(self):
        """
        Method that checks for data consistency.
        """
        self._check_data_digest_overlap_psms()
        self._check_data_digest_overlap_proteins()

    def _check_data_digest_overlap_psms(self):
        """
        Method that logs the overlap between the digested fasta file and the input files on the PSM level.
        """
        peptides = [x.stripped_peptide for x in self.main_data_form]
        peptides_in_digest = set(self.digest.peptide_to_protein_dictionary.keys())
        peptides_from_search_in_digest = [x for x in peptides if x in peptides_in_digest]
        percentage = float(len(set(peptides))) / float(len(set(peptides_from_search_in_digest)))
        logger.info("{} PSMs identified from input files".format(len(peptides)))
        logger.info(
            "{} PSMs identified from input files that are also present in database digestion".format(
                len(peptides_from_search_in_digest)
            )
        )
        logger.info(
            "{}; ratio of PSMs identified from input files to those that are present in the search"
            " and in the database digestion".format(percentage)
        )

    def _check_data_digest_overlap_proteins(self):
        """
        Method that logs the overlap between the digested fasta file and the input files on the Protein level.
        """
        proteins = [x.possible_proteins for x in self.main_data_form]
        flat_proteins = set([item for sublist in proteins for item in sublist])
        proteins_in_digest = set(self.digest.protein_to_peptide_dictionary.keys())
        proteins_from_search_in_digest = [x for x in flat_proteins if x in proteins_in_digest]
        percentage = float(len(flat_proteins)) / float(len(proteins_from_search_in_digest))
        logger.info("{} proteins identified from input files".format(len(flat_proteins)))
        logger.info(
            "{} proteins identified from input files that are also present in database digestion".format(
                len(proteins_from_search_in_digest)
            )
        )
        logger.info(
            "{}; ratio of proteins identified from input files that are also present in database digestion".format(
                percentage
            )
        )

    def _check_target_decoy_split(self):
        """
        Method that logs the number of target and decoy proteins from the digest.
        """
        # Check the number of targets vs the number of decoys from the digest
        targets = [
            x
            for x in self.digest.protein_to_peptide_dictionary.keys()
            if self.parameter_file_object.decoy_symbol not in x
        ]
        decoys = [
            x for x in self.digest.protein_to_peptide_dictionary.keys() if self.parameter_file_object.decoy_symbol in x
        ]
        ratio = float(len(targets)) / float(len(decoys))
        logger.info("Number of Target Proteins in Digest: {}".format(len(targets)))
        logger.info("Number of Decoy Proteins in Digest: {}".format(len(decoys)))
        logger.info("Ratio of Targets Proteins to Decoy Proteins: {}".format(ratio))

    def _validate_decoys_from_data(self):
        """
        Method that checks to make sure that target and decoy proteins exist in the data files.
        """
        # Check to see if we find decoys from our input files
        proteins = [x.possible_proteins for x in self.main_data_form]
        flat_proteins = set([item for sublist in proteins for item in sublist])
        targets = [x for x in flat_proteins if self.parameter_file_object.decoy_symbol not in x]
        decoys = [x for x in flat_proteins if self.parameter_file_object.decoy_symbol in x]
        logger.info("Number of Target Proteins in Data Files: {}".format(len(targets)))
        logger.info("Number of Decoy Proteins in Data Files: {}".format(len(decoys)))

    def _validate_isoform_from_data(self):
        """
        Method that validates whether or not isoforms are able to be identified in the data files.
        """
        # Check to see if we find any proteins with isoform info in name in our input files
        proteins = [x.possible_proteins for x in self.main_data_form]
        flat_proteins = set([item for sublist in proteins for item in sublist])
        if self.parameter_file_object.isoform_symbol:
            non_iso = [x for x in flat_proteins if self.parameter_file_object.isoform_symbol not in x]

        else:
            non_iso = [x for x in flat_proteins]

        if self.parameter_file_object.isoform_symbol:
            iso = [x for x in flat_proteins if self.parameter_file_object.isoform_symbol in x]

        else:
            iso = []
        logger.info("Number of Non Isoform Labeled Proteins in Data Files: {}".format(len(non_iso)))
        logger.info("Number of Isoform Labeled Proteins in Data Files: {}".format(len(iso)))

    def _validate_reviewed_v_unreviewed(self):
        """
        Method that logs whether or not we can distinguish from reviewed and unreviewd protein identifiers
        in the digest.
        """
        # Check to see if we get reviewed prots in digest...
        reviewed_proteins = len(self.digest.swiss_prot_protein_set)
        proteins_in_digest = len(set(self.digest.protein_to_peptide_dictionary.keys()))
        unreviewed_proteins = proteins_in_digest - reviewed_proteins
        logger.info("Number of Total Proteins in from Digest: {}".format(proteins_in_digest))
        logger.info("Number of Reviewed Proteins in from Digest: {}".format(reviewed_proteins))
        logger.info("Number of Unreviewed Proteins in from Digest: {}".format(unreviewed_proteins))

    @classmethod
    def sort_protein_strings(cls, protein_string_list, sp_proteins, decoy_symbol):
        """
        Method that sorts protein strings in the following order: Target Reviewed, Decoy Reviewed, Target Unreviewed,
         Decoy Unreviewed.

        Args:
            protein_string_list (list): List of Protein Strings.
            sp_proteins (set): Set of Reviewed Protein Strings.
            decoy_symbol (str): Symbol to denote a decoy protein identifier IE "##".

        Returns:
            list: List of sorted protein strings.

        Example:
            >>> list_of_group_objects = datastore.DataStore.sort_protein_strings(
            >>>     protein_string_list=protein_string_list, sp_proteins=sp_proteins, decoy_symbol="##"
            >>> )
        """

        our_target_sp_proteins = sorted([x for x in protein_string_list if x in sp_proteins and decoy_symbol not in x])
        our_decoy_sp_proteins = sorted([x for x in protein_string_list if x in sp_proteins and decoy_symbol in x])

        our_target_tr_proteins = sorted(
            [x for x in protein_string_list if x not in sp_proteins and decoy_symbol not in x]
        )
        our_decoy_tr_proteins = sorted([x for x in protein_string_list if x not in sp_proteins and decoy_symbol in x])

        identifiers_sorted = (
            our_target_sp_proteins + our_decoy_sp_proteins + our_target_tr_proteins + our_decoy_tr_proteins
        )

        return identifiers_sorted

    def input_has_q(self):
        """
        Method that checks to see if the input data has q values.
        """
        len_q = len([x.qvalue for x in self.main_data_form if x.qvalue])
        len_all = len(self.main_data_form)
        if len_q == len_all:
            status = True
            logger.info("Input has Q value; Can restrict by Q value")
        else:
            status = False
            logger.warning("Input does not have Q value; Cannot restrict by Q value")

        return status

    def input_has_pep(self):
        """
        Method that checks to see if the input data has pep values.
        """
        len_pep = len([x.pepvalue for x in self.main_data_form if x.pepvalue])
        len_all = len(self.main_data_form)
        if len_pep == len_all:
            status = True
            logger.info("Input has Pep value; Can restrict by Pep value")
        else:
            status = False
            logger.warning("Input does not have Pep value; Cannot restrict by Pep value")

        return status

    def input_has_custom(self):
        """
        Method that checks to see if the input data has custom score values.
        """
        len_c = len([x.custom_score for x in self.main_data_form if x.custom_score])
        len_all = len(self.main_data_form)
        if len_c == len_all:
            status = True
            logger.info("Input has Custom value; Can restrict by Custom value")

        else:
            status = False
            logger.warning("Input does not have Custom value; Cannot restrict by Custom value")

        return status

    def get_protein_objects(self, false_discovery_rate=None, fdr_restricted=False):
        """
        Method retrieves protein objects. Either retrieves FDR restricted list of protien objects,
        or retrieves all objects.

        Args:
            fdr_restricted (bool): True/False on whether to restrict the list of objects based on FDR.

        Returns:
            list: List of scored [ProteinGroup][pyproteininference.physical.ProteinGroup] objects
                that have been grouped and sorted.

        """
        if not false_discovery_rate:
            false_discovery_rate = self.parameter_file_object.fdr
        if fdr_restricted:
            protein_objects = [x.proteins for x in self.protein_group_objects if x.q_value <= false_discovery_rate]
        else:
            protein_objects = self.grouped_scored_proteins

        return protein_objects

    def _init_validate(self, reader):
        """
        Internal Method that checks to make sure the reader object is properly loaded and validated.
        """
        if reader.psms:
            self.main_data_form = reader.psms  # Unrestricted PSM data
            self.restricted_peptides = [x.non_flanking_peptide for x in self.main_data_form]
        else:
            raise ValueError(
                "Psms variable from Reader object is either empty or does not exist. "
                "Make sure your files contain proper data and that you run the 'read_psms' "
                "method on your Reader object."
            )

    def _validate_main_data_form(self):
        """
        Internal Method that checks to make sure the Main data has been defined to run DataStore methods.
        """
        if self.main_data_form:
            pass
        else:
            raise ValueError(
                "Main Data is not defined, thus method cannot be ran. Please make sure PSM data is properly"
                " loaded from the Reader object"
            )

    def _validate_main_data_restricted(self):
        """
        Internal Method that checks to make sure the Main data Restricted has been defined to run DataStore methods.
        """
        if self.main_data_restricted:
            pass
        else:
            raise ValueError(
                "Main Data Restricted is not defined, thus method cannot be ran. Please make sure PSM data is properly"
                " loaded from the Reader object and make sure to run DataStore method 'restrict_psm_data'."
            )

    def _validate_scored_proteins(self):
        """
        Internal Method that checks to make sure that proteins have been scored to run certain subsequent methods.
        """
        if self.picked_proteins_scored or self.scored_proteins:
            pass
        else:
            raise ValueError(
                "Proteins have not been scored, Please initialize a Score object and run a score method with"
                " 'score_psms' instance method."
            )

    def _validate_scoring_input(self):
        """
        Internal Method that checks to make sure that Scoring Input has been created to be able to run scoring methods.
        """
        if self.scoring_input:
            pass
        else:
            raise ValueError(
                "Scoring input has not been created, Please run 'create_scoring_input' method from the DataStore "
                "object to continue."
            )

    def _validate_protein_group_objects(self):
        """
        Internal Method that checks to make sure inference has been run before proceeding.
        """
        if self.protein_group_objects and self.grouped_scored_proteins:
            pass
        else:
            raise ValueError(
                "Either 'protein_group_objects' or 'grouped_scored_proteins' or both DataStore variables are undefined."
                " Please make sure you run an inference method from the Inference class before proceeding."
            )

    def generate_fdr_vs_target_hits(self, fdr_max=0.2):
        """
        Method for calculating FDR vs number of Target Proteins.

        Args:
            fdr_max (float): The maximum false discovery rate to calculate target hits for.
                Will stop once fdr_max is reached.

        Returns:
            list: List of lists of: (FDR, Number of Target Hits). Ordered by increasing number of Target Hits.

        """
        fdr_vs_count = []
        count_list = []
        for pg in self.protein_group_objects:
            if self.decoy_symbol not in pg.proteins[0].identifier:
                count_list.append(pg)
            fdr_vs_count.append([pg.q_value, len(count_list)])

        fdr_vs_count = [x for x in fdr_vs_count if x[0] < fdr_max]

        return fdr_vs_count

    def recover_mapping(self):
        logger.info("Recovering Proteins that exist in the input files but not in the database digest.")
        all_psms = self.get_psm_data()
        proteins = [x.possible_proteins for x in all_psms]
        flat_proteins = [item for sublist in proteins for item in sublist]

        missing_prots = []
        for prot in flat_proteins:
            try:
                self.digest.protein_to_peptide_dictionary[prot]
            except KeyError:
                missing_prots.append(prot)

                psm_data = self.get_psm_data()
                peptides = [x.stripped_peptide for x in psm_data if prot in x.possible_proteins]
                for pep in peptides:
                    self.digest.peptide_to_protein_dictionary.setdefault(pep, set()).add(prot)
                    self.digest.protein_to_peptide_dictionary.setdefault(prot, set()).add(pep)
        if missing_prots:
            logger.info(
                "{} proteins not found in mapping objects, please double check that your database"
                " provided is accurate for the given input data.".format(len(missing_prots))
            )
        else:
            logger.info("No missing proteins in the mapping objects.")



from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_exclusion.yaml")


class TestMappingRecovery(TestCase):
    def test_mapping_recovery(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        self.assertEqual(len(data.main_data_restricted), 26)

        del data.digest.protein_to_peptide_dictionary["Q96BA7_HUMAN|Q96BA7"]
        del data.digest.protein_to_peptide_dictionary["B3KX72_HUMAN|B3KX72"]

        for pep in list(data.digest.peptide_to_protein_dictionary.keys()):
            if "Q96BA7_HUMAN|Q96BA7" in data.digest.peptide_to_protein_dictionary[pep]:
                data.digest.peptide_to_protein_dictionary[pep].remove("Q96BA7_HUMAN|Q96BA7")
            if "B3KX72_HUMAN|B3KX72" in data.digest.peptide_to_protein_dictionary[pep]:
                data.digest.peptide_to_protein_dictionary[pep].remove("B3KX72_HUMAN|B3KX72")

        # Try to run exclusion with proteins missing from the database digest but existing in the main psm data
        with self.assertRaises(KeyError):
            data.create_scoring_input()
            data.exclude_non_distinguishing_peptides()

        # Recover the proteins and check the digests
        data.recover_mapping()

        self.assertIn("Q96BA7_HUMAN|Q96BA7", list(data.digest.protein_to_peptide_dictionary.keys()))
        self.assertIn("B3KX72_HUMAN|B3KX72", list(data.digest.protein_to_peptide_dictionary.keys()))
        self.assertIn(
            "Q96BA7_HUMAN|Q96BA7", data.digest.peptide_to_protein_dictionary['LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR']
        )
        self.assertIn(
            "B3KX72_HUMAN|B3KX72", data.digest.peptide_to_protein_dictionary['LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR']
        )
        self.assertIn("B3KX72_HUMAN|B3KX72", data.digest.peptide_to_protein_dictionary['AEGGGGGGRPGAPAAGDGK'])

        data.create_scoring_input()
        data.exclude_non_distinguishing_peptides()



import os
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE_ORIGINAL = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TEST_DATABASE_MISSING_PEP = resource_filename("pyproteininference", "../tests/data/test_database_missing_peptide.fasta")
TEST_DATABASE_MISSING_PROT = resource_filename(
    "pyproteininference", "../tests/data/test_database_missing_protein.fasta"
)
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')
for sub_dir in ["leads", "all", "peptides", "psms", "psm_ids"]:
    if not os.path.exists(os.path.join(OUTPUT_DIR, sub_dir)):
        os.makedirs(os.path.join(OUTPUT_DIR, sub_dir))

LEAD_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/leads/test_inclusion_q_value_leads_ml_posterior_error_prob.csv",
)
ALL_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/all/test_inclusion_q_value_all_ml_posterior_error_prob.csv",
)
PEPTIDE_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/peptides/test_inclusion_q_value_leads_peptides_ml_posterior_error_prob.csv",
)
PSM_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/psms/test_inclusion_q_value_leads_psms_ml_posterior_error_prob.csv",
)
PSM_ID_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/psm_ids/test_inclusion_q_value_leads_psm_ids_ml_posterior_error_prob.csv",
)

IDENTIFIER_INDEX = 0
SCORE_INDEX = 1
Q_VALUE_INDEX = 2
GROUP_ID_INDEX = 5
PEPTIDES_INDEX = 6


class TestFaultyDatabasePipeline(TestCase):
    def test_faulty_database_pipeline(self):

        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # Perform digest on each database
        digest_normal = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE_ORIGINAL,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest_normal.digest_fasta_database()

        digest_missing_pep = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE_MISSING_PEP,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest_missing_pep.digest_fasta_database()

        digest_missing_prot = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE_MISSING_PROT,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest_missing_prot.digest_fasta_database()

        # Read PSM data with each of the digests with varying degrees of missingness
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest_normal,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        pep_and_prot_data_miss_pep = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest_missing_pep,
        )
        pep_and_prot_data_miss_pep.read_psms()

        self.assertEqual(len(pep_and_prot_data_miss_pep.psms), 27)

        pep_and_prot_data_miss_prot = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest_missing_prot,
        )
        pep_and_prot_data_miss_prot.read_psms()

        self.assertEqual(len(pep_and_prot_data_miss_prot.psms), 27)

        # Load the digests and reader objects into datastore objects
        data_normal = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest_normal)

        data_miss_pep = pyproteininference.datastore.DataStore(pep_and_prot_data_miss_pep, digest=digest_missing_pep)

        data_miss_prot = pyproteininference.datastore.DataStore(pep_and_prot_data_miss_prot, digest=digest_missing_prot)

        for i in range(len(data_normal.main_data_form)):
            self.assertEqual(
                data_normal.main_data_form[0].identifier,
                data_miss_pep.main_data_form[0].identifier,
            )
            self.assertEqual(
                data_normal.main_data_form[0].non_flanking_peptide,
                data_miss_pep.main_data_form[0].non_flanking_peptide,
            )
            self.assertEqual(
                data_normal.main_data_form[0].pepvalue,
                data_miss_pep.main_data_form[0].pepvalue,
            )
            self.assertEqual(
                data_normal.main_data_form[0].percscore,
                data_miss_pep.main_data_form[0].percscore,
            )
            self.assertListEqual(
                data_normal.main_data_form[0].possible_proteins,
                data_miss_pep.main_data_form[0].possible_proteins,
            )
            self.assertEqual(
                data_normal.main_data_form[0].psm_id,
                data_miss_pep.main_data_form[0].psm_id,
            )
            self.assertEqual(
                data_normal.main_data_form[0].qvalue,
                data_miss_pep.main_data_form[0].qvalue,
            )
            self.assertEqual(
                data_normal.main_data_form[0].stripped_peptide,
                data_miss_pep.main_data_form[0].stripped_peptide,
            )

        for i in range(len(data_normal.main_data_form)):
            self.assertEqual(
                data_normal.main_data_form[0].identifier,
                data_miss_prot.main_data_form[0].identifier,
            )
            self.assertEqual(
                data_normal.main_data_form[0].non_flanking_peptide,
                data_miss_prot.main_data_form[0].non_flanking_peptide,
            )
            self.assertEqual(
                data_normal.main_data_form[0].pepvalue,
                data_miss_prot.main_data_form[0].pepvalue,
            )
            self.assertEqual(
                data_normal.main_data_form[0].percscore,
                data_miss_prot.main_data_form[0].percscore,
            )
            self.assertListEqual(
                data_normal.main_data_form[0].possible_proteins,
                data_miss_prot.main_data_form[0].possible_proteins,
            )
            self.assertEqual(
                data_normal.main_data_form[0].psm_id,
                data_miss_prot.main_data_form[0].psm_id,
            )
            self.assertEqual(
                data_normal.main_data_form[0].qvalue,
                data_miss_prot.main_data_form[0].qvalue,
            )
            self.assertEqual(
                data_normal.main_data_form[0].stripped_peptide,
                data_miss_prot.main_data_form[0].stripped_peptide,
            )

        stripped_peptides = [x.stripped_peptide for x in data_normal.main_data_form]
        possible_proteins = [x.possible_proteins for x in data_normal.main_data_form]

        proteins = set([y for x in possible_proteins for y in x])

        # Loop over all peptides from the normal data and make sure the peptide a key in peptide_to_protein dict
        for peps in stripped_peptides:
            self.assertIn(peps, digest_normal.peptide_to_protein_dictionary.keys())
            self.assertIn(peps, digest_missing_pep.peptide_to_protein_dictionary.keys())
            self.assertIn(peps, digest_missing_prot.peptide_to_protein_dictionary.keys())
            self.assertSetEqual(
                digest_normal.peptide_to_protein_dictionary[peps],
                digest_missing_pep.peptide_to_protein_dictionary[peps],
            )
            self.assertSetEqual(
                digest_normal.peptide_to_protein_dictionary[peps],
                digest_missing_prot.peptide_to_protein_dictionary[peps],
            )

        for prots in proteins:
            self.assertIn(prots, digest_normal.protein_to_peptide_dictionary.keys())
            self.assertIn(prots, digest_missing_pep.protein_to_peptide_dictionary.keys())
            self.assertIn(prots, digest_missing_prot.protein_to_peptide_dictionary.keys())

        # Check to make sure datastore dictionaries are equal
        self.assertDictEqual(
            data_normal.peptide_to_protein_dictionary(),
            data_miss_pep.peptide_to_protein_dictionary(),
        )
        self.assertDictEqual(
            data_normal.peptide_to_protein_dictionary(),
            data_miss_prot.peptide_to_protein_dictionary(),
        )

        self.assertDictEqual(
            data_normal.protein_to_peptide_dictionary(),
            data_miss_pep.protein_to_peptide_dictionary(),
        )
        self.assertDictEqual(
            data_normal.protein_to_peptide_dictionary(),
            data_miss_prot.protein_to_peptide_dictionary(),
        )



import csv
import logging
import os
import sys

logger = logging.getLogger(__name__)

# set up our logger
logging.basicConfig(
    stream=sys.stderr,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class Export(object):
    """
    Class that handles exporting protein inference results to filesystem as csv files.

    Attributes:
        data (DataStore): [DataStore object][pyproteininference.datastore.DataStore].
        filepath (str): Path to file to be written.

    """

    EXPORT_LEADS = "leads"
    EXPORT_ALL = "all"
    EXPORT_COMMA_SEP = "comma_sep"
    EXPORT_Q_VALUE_COMMA_SEP = "q_value_comma_sep"
    EXPORT_Q_VALUE = "q_value"
    EXPORT_Q_VALUE_ALL = "q_value_all"
    EXPORT_PEPTIDES = "peptides"
    EXPORT_PSMS = "psms"
    EXPORT_PSM_IDS = "psm_ids"
    EXPORT_LONG = "long"

    EXPORT_TYPES = [
        EXPORT_LEADS,
        EXPORT_ALL,
        EXPORT_COMMA_SEP,
        EXPORT_Q_VALUE_COMMA_SEP,
        EXPORT_Q_VALUE,
        EXPORT_Q_VALUE_ALL,
        EXPORT_PEPTIDES,
        EXPORT_PSMS,
        EXPORT_PSM_IDS,
        EXPORT_LONG,
    ]

    def __init__(self, data):
        """
        Initialization method for the Export class.

        Args:
            data (DataStore): [DataStore object][pyproteininference.datastore.DataStore].

        Example:
            >>> export = pyproteininference.export.Export(data=data)

        """
        self.data = data
        self.filepath = None

    def export_to_csv(self, output_filename=None, directory=None, export_type="q_value"):
        """
        Method that dispatches to one of the many export methods given an export_type input.

        filepath is determined based on directory arg and information from
        [DataStore object][pyproteininference.datastore.DataStore].

        This method sets the `filepath` variable.

        Args:
            output_filename (str): Filepath to write to. If set as None will auto generate filename and
                will write to directory variable.
            directory (str): Directory to write the result file to. If None, will write to current working directory.
            export_type (str): Must be a value in `EXPORT_TYPES` and determines the output format.

        Example:
            >>> export = pyproteininference.export.Export(data=data)
            >>> export.export_to_csv(output_filename=None, directory="/path/to/output/dir/", export_type="psms")

        """

        if not directory:
            directory = os.getcwd()

        data = self.data
        tag = data.parameter_file_object.tag

        if self.EXPORT_LEADS == export_type:
            filename = "{}_leads_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_leads_restricted(filename_out=complete_filepath)

        elif self.EXPORT_ALL == export_type:
            filename = "{}_all_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_all_restricted(complete_filepath)

        elif self.EXPORT_COMMA_SEP == export_type:
            filename = "{}_comma_sep_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_comma_sep_restricted(complete_filepath)

        elif self.EXPORT_Q_VALUE_COMMA_SEP == export_type:
            filename = "{}_q_value_comma_sep_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_comma_sep(complete_filepath)

        elif self.EXPORT_Q_VALUE == export_type:
            filename = "{}_q_value_leads_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_leads(complete_filepath)

        elif self.EXPORT_Q_VALUE_ALL == export_type:
            filename = "{}_q_value_all_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_all(complete_filepath)

        elif self.EXPORT_PEPTIDES == export_type:
            filename = "{}_q_value_leads_peptides_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_leads_peptides(complete_filepath)

        elif self.EXPORT_PSMS == export_type:
            filename = "{}_q_value_leads_psms_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_leads_psms(complete_filepath)

        elif self.EXPORT_PSM_IDS == export_type:
            filename = "{}_q_value_leads_psm_ids_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_leads_psm_ids(complete_filepath)

        elif self.EXPORT_LONG == export_type:
            filename = "{}_q_value_long_{}_{}.csv".format(tag, data.short_protein_score, data.psm_score)
            complete_filepath = os.path.join(directory, filename)
            if output_filename:
                complete_filepath = output_filename
            logger.info("Exporting Protein Inference Data to File: {}".format(complete_filepath))
            self.csv_export_q_value_leads_long(complete_filepath)

        else:
            complete_filepath = "protein_inference_results.csv"

        self.filepath = complete_filepath

    def csv_export_all_restricted(self, filename_out):
        """
        Method that outputs a subset of the passing proteins based on FDR.

        This method returns a non-square CSV file.

        Args:
            filename_out (str): Filename for the data to be written to

        """
        protein_objects = self.data.get_protein_objects(fdr_restricted=True)
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in protein_objects:
            for prots in groups:
                protein_export_list.append([prots.identifier])
                protein_export_list[-1].append(prots.score)
                protein_export_list[-1].append(prots.num_peptides)
                if prots.reviewed:
                    protein_export_list[-1].append("Reviewed")
                else:
                    protein_export_list[-1].append("Unreviewed")
                protein_export_list[-1].append(prots.group_identification)
                for peps in prots.peptides:
                    protein_export_list[-1].append(peps)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_leads_restricted(self, filename_out):
        """
        Method that outputs a subset of the passing proteins based on FDR.
        Only Proteins that pass FDR will be output and only Lead proteins will be output

        This method returns a non-square CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.

        """
        protein_objects = self.data.get_protein_objects(fdr_restricted=True)
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in protein_objects:
            protein_export_list.append([groups[0].identifier])
            protein_export_list[-1].append(groups[0].score)
            protein_export_list[-1].append(groups[0].num_peptides)
            if groups[0].reviewed:
                protein_export_list[-1].append("Reviewed")
            else:
                protein_export_list[-1].append("Unreviewed")
            protein_export_list[-1].append(groups[0].group_identification)
            for peps in sorted(groups[0].peptides):
                protein_export_list[-1].append(peps)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_comma_sep_restricted(self, filename_out):
        """
        Method that outputs a subset of the passing proteins based on FDR.
        Only Proteins that pass FDR will be output and only Lead proteins will be output.
        Proteins in the groups of lead proteins will also be output in the same row as the lead.

        This method returns a non-square CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.

        """
        protein_objects = self.data.get_protein_objects(fdr_restricted=True)
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Other_Potential_Identifiers",
            ]
        ]
        for groups in protein_objects:
            for prots in groups:
                if prots == groups[0]:
                    protein_export_list.append([prots.identifier])
                    protein_export_list[-1].append(prots.score)
                    protein_export_list[-1].append(prots.num_peptides)
                    if prots.reviewed:
                        protein_export_list[-1].append("Reviewed")
                    else:
                        protein_export_list[-1].append("Unreviewed")
                    protein_export_list[-1].append(prots.group_identification)
                else:
                    protein_export_list[-1].append(prots.identifier)
        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_leads(self, filename_out):
        """
        Method that outputs all lead proteins with Q values.

        This method returns a non-square CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.

        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            lead_protein = groups.proteins[0]
            protein_export_list.append([lead_protein.identifier])
            protein_export_list[-1].append(lead_protein.score)
            protein_export_list[-1].append(groups.q_value)
            protein_export_list[-1].append(lead_protein.num_peptides)
            if lead_protein.reviewed:
                protein_export_list[-1].append("Reviewed")
            else:
                protein_export_list[-1].append("Unreviewed")
            protein_export_list[-1].append(groups.number_id)
            peptides = lead_protein.peptides
            for peps in sorted(peptides):
                protein_export_list[-1].append(peps)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_comma_sep(self, filename_out):
        """
        Method that outputs all lead proteins with Q values.
        Proteins in the groups of lead proteins will also be output in the same row as the lead.

        This method returns a non-square CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.

        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Other_Potential_Identifiers",
            ]
        ]
        for groups in self.data.protein_group_objects:
            lead_protein = groups.proteins[0]
            protein_export_list.append([lead_protein.identifier])
            protein_export_list[-1].append(lead_protein.score)
            protein_export_list[-1].append(groups.q_value)
            protein_export_list[-1].append(lead_protein.num_peptides)
            if lead_protein.reviewed:
                protein_export_list[-1].append("Reviewed")
            else:
                protein_export_list[-1].append("Unreviewed")
            protein_export_list[-1].append(groups.number_id)
            for other_prots in groups.proteins[1:]:
                protein_export_list[-1].append(other_prots.identifier)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_all(self, filename_out):
        """
        Method that outputs all proteins with Q values.
        Non Lead proteins are also output - entire group gets output.
        Proteins in the groups of lead proteins will also be output in the same row as the lead.

        This method returns a non-square CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.

        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            for proteins in groups.proteins:
                protein_export_list.append([proteins.identifier])
                protein_export_list[-1].append(proteins.score)
                protein_export_list[-1].append(groups.q_value)
                protein_export_list[-1].append(proteins.num_peptides)
                if proteins.reviewed:
                    protein_export_list[-1].append("Reviewed")
                else:
                    protein_export_list[-1].append("Unreviewed")
                protein_export_list[-1].append(groups.number_id)
                for peps in sorted(proteins.peptides):
                    protein_export_list[-1].append(peps)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_all_proteologic(self, filename_out):
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            for proteins in groups.proteins:
                protein_export_list.append([proteins.identifier])
                protein_export_list[-1].append(proteins.score)
                protein_export_list[-1].append(groups.q_value)
                protein_export_list[-1].append(proteins.num_peptides)
                if proteins.reviewed:
                    protein_export_list[-1].append("Reviewed")
                else:
                    protein_export_list[-1].append("Unreviewed")
                protein_export_list[-1].append(groups.number_id)
                for peps in sorted(proteins.peptides):
                    protein_export_list[-1].append(peps)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_leads_long(self, filename_out):
        """
        Method that outputs all lead proteins with Q values.

        This method returns a long formatted result file with one peptide on each row.

        Args:
            filename_out (str): Filename for the data to be written to.

        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            lead_protein = groups.proteins[0]
            for peps in sorted(lead_protein.peptides):
                protein_export_list.append([lead_protein.identifier])
                protein_export_list[-1].append(lead_protein.score)
                protein_export_list[-1].append(groups.q_value)
                protein_export_list[-1].append(lead_protein.num_peptides)
                if lead_protein.reviewed:
                    protein_export_list[-1].append("Reviewed")
                else:
                    protein_export_list[-1].append("Unreviewed")
                protein_export_list[-1].append(groups.number_id)
                protein_export_list[-1].append(peps)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_leads_peptides(self, filename_out, peptide_delimiter=" "):
        """
        Method that outputs all lead proteins with Q values in rectangular format.
        This method outputs unique peptides per protein.

        This method returns a rectangular CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.
            peptide_delimiter (str): String to separate peptides by in the "Peptides" column of the csv file
        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            lead_protein = groups.proteins[0]
            protein_export_list.append([lead_protein.identifier])
            protein_export_list[-1].append(lead_protein.score)
            protein_export_list[-1].append(groups.q_value)
            protein_export_list[-1].append(lead_protein.num_peptides)
            if lead_protein.reviewed:
                protein_export_list[-1].append("Reviewed")
            else:
                protein_export_list[-1].append("Unreviewed")
            protein_export_list[-1].append(groups.number_id)
            peptides = peptide_delimiter.join(list(sorted(lead_protein.peptides)))
            protein_export_list[-1].append(peptides)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_leads_psms(self, filename_out, peptide_delimiter=" "):
        """
        Method that outputs all lead proteins with Q values in rectangular format.
        This method outputs all PSMs for the protein not just unique peptide identifiers.

        This method returns a rectangular CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.
            peptide_delimiter (str): String to separate peptides by in the "Peptides" column of the csv file.
        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            lead_protein = groups.proteins[0]
            protein_export_list.append([lead_protein.identifier])
            protein_export_list[-1].append(lead_protein.score)
            protein_export_list[-1].append(groups.q_value)
            protein_export_list[-1].append(lead_protein.num_peptides)
            if lead_protein.reviewed:
                protein_export_list[-1].append("Reviewed")
            else:
                protein_export_list[-1].append("Unreviewed")
            protein_export_list[-1].append(groups.number_id)
            psms = peptide_delimiter.join(sorted([x.non_flanking_peptide for x in lead_protein.psms]))
            protein_export_list[-1].append(psms)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)

    def csv_export_q_value_leads_psm_ids(self, filename_out, peptide_delimiter=" "):
        """
        Method that outputs all lead proteins with Q values in rectangular format.
        Psms are output as the psm_id value. So sequence information is not output.

        This method returns a rectangular CSV file.

        Args:
            filename_out (str): Filename for the data to be written to.
            peptide_delimiter (str): String to separate psm_ids by in the "Peptides" column of the csv file.
        """
        protein_export_list = [
            [
                "Protein",
                "Score",
                "Q_Value",
                "Number_of_Peptides",
                "Identifier_Type",
                "GroupID",
                "Peptides",
            ]
        ]
        for groups in self.data.protein_group_objects:
            lead_protein = groups.proteins[0]
            protein_export_list.append([lead_protein.identifier])
            protein_export_list[-1].append(lead_protein.score)
            protein_export_list[-1].append(groups.q_value)
            protein_export_list[-1].append(lead_protein.num_peptides)
            if lead_protein.reviewed:
                protein_export_list[-1].append("Reviewed")
            else:
                protein_export_list[-1].append("Unreviewed")
            protein_export_list[-1].append(groups.number_id)
            psms = peptide_delimiter.join(sorted(lead_protein.get_psm_ids()))
            protein_export_list[-1].append(psms)

        with open(filename_out, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(protein_export_list)



import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')

TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")

temp_dir = tempfile.gettempdir()


class TestScoreMethods(TestCase):
    def test_scoring_methods(self):
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = pyproteininference.parameters.ProteinInferenceParameter(
            yaml_param_filepath=PARAMETER_FILE
        )

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = pyproteininference.in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            combined_files=None,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method="best_peptide_per_protein")
        self.assertEqual(data.scored_proteins[0].score, 3.5e-06)

        score.score_psms(score_method="iterative_downweighted_log")
        self.assertEqual(data.scored_proteins[0].score, 64.16418132258529)

        score.score_psms(score_method="multiplicative_log")
        self.assertEqual(data.scored_proteins[0].score, 82.89306334778564)

        score.score_psms(score_method="downweighted_multiplicative_log")
        self.assertEqual(data.scored_proteins[0].score, 38.42870136439401)

        score.score_psms(score_method="downweighted_version2")
        self.assertEqual(data.scored_proteins[0].score, 29.723342219506584)

        score.score_psms(score_method="top_two_combined")
        self.assertEqual(data.scored_proteins[0].score, 25.523842819277114)

        score.score_psms(score_method="geometric_mean")
        self.assertEqual(data.scored_proteins[0].score, 11.790572089099232)



from unittest import TestCase

from pyproteininference.physical import Psm


class TestLoadPeptideParsing(TestCase):
    def test_peptide_parsing(self):

        # GFY Sequence Flanking
        peptide1 = "R.ALASLM*TYK.-"
        peptide1_sf = Psm.split_peptide(peptide_string=peptide1)
        self.assertEqual("ALASLM*TYK", peptide1_sf)

        peptide1_no_mods = Psm.remove_peptide_mods(peptide_string=peptide1_sf)
        self.assertEqual("ALASLMTYK", peptide1_no_mods)

        # GFY Sequence no Flanking
        peptide2 = "ALASLM*TYK"
        peptide2_sf = Psm.split_peptide(peptide_string=peptide2)
        self.assertEqual("ALASLM*TYK", peptide2_sf)

        peptide2_no_mods = Psm.remove_peptide_mods(peptide_string=peptide2_sf)
        self.assertEqual("ALASLMTYK", peptide2_no_mods)

        # Max Quant Sequence Flanking _ and string mod
        peptide3 = "_ALASLM(ox)TYK_"
        peptide3_sf = Psm.split_peptide(peptide_string=peptide3, delimiter="_")
        self.assertEqual("ALASLM(ox)TYK", peptide3_sf)

        peptide3_no_mods = Psm.remove_peptide_mods(peptide_string=peptide3_sf)
        self.assertEqual("ALASLMTYK", peptide3_no_mods)

        # Max Quant with float mass mod
        peptide4 = "ALASLM(137.9)TYK"
        peptide4_sf = Psm.split_peptide(peptide_string=peptide4)
        self.assertEqual("ALASLM(137.9)TYK", peptide4_sf)

        peptide4_no_mods = Psm.remove_peptide_mods(peptide_string=peptide4_sf)
        self.assertEqual("ALASLMTYK", peptide4_no_mods)

        # Max Quant sequence with int mass mod
        peptide5 = "ALASLM(1)TYK"
        peptide5_sf = Psm.split_peptide(peptide_string=peptide5)
        self.assertEqual("ALASLM(1)TYK", peptide5_sf)

        peptide5_no_mods = Psm.remove_peptide_mods(peptide_string=peptide5_sf)
        self.assertEqual("ALASLMTYK", peptide5_no_mods)

        # Max Quant with float mass and flanking with "."
        peptide6 = "R.ALASLM(137.9)TYK.-"
        peptide6_sf = Psm.split_peptide(peptide_string=peptide6)
        self.assertEqual("ALASLM(137.9)TYK", peptide6_sf)

        peptide6_no_mods = Psm.remove_peptide_mods(peptide_string=peptide6_sf)
        self.assertEqual("ALASLMTYK", peptide6_no_mods)

        # Comet Pepxml with brackets and int mod and n term
        peptide7 = "n[230]M[147]PVLSAR"
        peptide7_sf = Psm.split_peptide(peptide_string=peptide7)
        self.assertEqual("n[230]M[147]PVLSAR", peptide7_sf)

        peptide7_no_mods = Psm.remove_peptide_mods(peptide_string=peptide7_sf)
        self.assertEqual("MPVLSAR", peptide7_no_mods)

        # Blank sequence
        peptide8 = "MPVLSAR"
        peptide8_sf = Psm.split_peptide(peptide_string=peptide8)
        self.assertEqual("MPVLSAR", peptide8_sf)

        peptide8_no_mods = Psm.remove_peptide_mods(peptide_string=peptide8_sf)
        self.assertEqual("MPVLSAR", peptide8_no_mods)

        # Blank sequence Mascot
        peptide9 = "MLSVGLGFLR"
        peptide9_sf = Psm.split_peptide(peptide_string=peptide9)
        self.assertEqual("MLSVGLGFLR", peptide9_sf)

        peptide9_no_mods = Psm.remove_peptide_mods(peptide_string=peptide9_sf)
        self.assertEqual("MLSVGLGFLR", peptide9_no_mods)

        # Peaks, no flanking and float mod within parentheses
        peptide10 = "VM(+15.99)YKFLTV"
        peptide10_sf = Psm.split_peptide(peptide_string=peptide10)
        self.assertEqual("VM(+15.99)YKFLTV", peptide10_sf)

        peptide10_no_mods = Psm.remove_peptide_mods(peptide_string=peptide10_sf)
        self.assertEqual("VMYKFLTV", peptide10_no_mods)

        # Peaks, with flanking and float mod within parentheses
        peptide11 = "R.VM(+15.99)YKFLTV.-"
        peptide11_sf = Psm.split_peptide(peptide_string=peptide11)
        self.assertEqual("VM(+15.99)YKFLTV", peptide11_sf)

        peptide11_no_mods = Psm.remove_peptide_mods(peptide_string=peptide11_sf)
        self.assertEqual("VMYKFLTV", peptide11_no_mods)

        # MSGF+ sequence with no flanking and no parentheses or brackets
        peptide12 = "NLANPTSVILASIQM+15.995LEYLGMADK"
        peptide12_sf = Psm.split_peptide(peptide_string=peptide12)
        self.assertEqual("NLANPTSVILASIQM+15.995LEYLGMADK", peptide12_sf)

        peptide12_no_mods = Psm.remove_peptide_mods(peptide_string=peptide12_sf)
        self.assertEqual("NLANPTSVILASIQMLEYLGMADK", peptide12_no_mods)

        # MSGF+ with flanking and parentheses and no brackets
        peptide13 = "K.NLANPTSVILASIQM+15.995LEYLGMADK.A"
        peptide13_sf = Psm.split_peptide(peptide_string=peptide13)
        self.assertEqual("NLANPTSVILASIQM+15.995LEYLGMADK", peptide13_sf)

        peptide13_no_mods = Psm.remove_peptide_mods(peptide_string=peptide13_sf)
        self.assertEqual("NLANPTSVILASIQMLEYLGMADK", peptide13_no_mods)



import os
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')
for sub_dir in ["leads", "all", "peptides", "psms", "psm_ids"]:
    if not os.path.exists(os.path.join(OUTPUT_DIR, sub_dir)):
        os.makedirs(os.path.join(OUTPUT_DIR, sub_dir))


class TestAltProteinRead(TestCase):
    def test_alt_protein_reader(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=True,
        )
        pep_and_prot_data.read_psms()

        possible_proteins = [x.possible_proteins for x in pep_and_prot_data.psms]

        true_possible_proteins = [
            ['RAF1_HUMAN|P04049'],
            ['RAF1_HUMAN|P04049'],
            ['ARAF_HUMAN|P10398', 'BRAF_HUMAN|P15056', 'RAF1_HUMAN|P04049'],
            ['ARAF_HUMAN|P10398', 'RAF1_HUMAN|P04049'],
            ['BRAF_HUMAN|P15056', 'RAF1_HUMAN|P04049'],
            ['ARAF_HUMAN|P10398', 'BRAF_HUMAN|P15056'],
            ['ARAF_HUMAN|P10398'],
            ['RAF1_HUMAN|P04049'],
            ['TCAF1_HUMAN|Q9Y4C2'],
            ['TCAF1_HUMAN|Q9Y4C2', 'TCAF1_HUMAN|Q9Y4C2-2'],
            ['HNRPU_HUMAN|Q00839', 'B3KX72_HUMAN|B3KX72', 'Q96BA7_HUMAN|Q96BA7'],
            ['HNRPU_HUMAN|Q00839', 'B3KX72_HUMAN|B3KX72', 'Q96BA7_HUMAN|Q96BA7'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['##TCAF1_HUMAN|##Q9Y4C2'],
            [
                '##TCAF2_HUMAN|##A6NFQ2',
                '##TCAF2_HUMAN|##A6NFQ2-2',
                '##TCAF2_HUMAN|##A6NFQ2-3',
            ],
            ['RAF1_HUMAN|P04049'],
        ]

        for i in range(len(possible_proteins)):
            self.assertSetEqual(set(possible_proteins[i]), set(true_possible_proteins[i]))

        pep_and_prot_data_no_append = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data_no_append.read_psms()

        possible_proteins_no_append = [x.possible_proteins for x in pep_and_prot_data_no_append.psms]

        true_possible_proteins_no_append = [
            ['RAF1_HUMAN|P04049'],
            ['RAF1_HUMAN|P04049'],
            ['ARAF_HUMAN|P10398', 'RAF1_HUMAN|P04049', 'BRAF_HUMAN|P15056'],
            ['ARAF_HUMAN|P10398', 'RAF1_HUMAN|P04049'],
            ['RAF1_HUMAN|P04049', 'BRAF_HUMAN|P15056'],
            ['ARAF_HUMAN|P10398', 'BRAF_HUMAN|P15056'],
            ['ARAF_HUMAN|P10398'],
            ['RAF1_HUMAN|P04049'],
            ['TCAF1_HUMAN|Q9Y4C2'],
            ['TCAF1_HUMAN|Q9Y4C2'],
            ['HNRPU_HUMAN|Q00839', 'B3KX72_HUMAN|B3KX72'],
            ['HNRPU_HUMAN|Q00839', 'B3KX72_HUMAN|B3KX72', 'Q96BA7_HUMAN|Q96BA7'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['##TCAF1_HUMAN|##Q9Y4C2'],
            ['##TCAF2_HUMAN|##A6NFQ2'],
            ['RAF1_HUMAN|P04049'],
        ]

        for i in range(len(possible_proteins_no_append)):
            self.assertSetEqual(
                set(possible_proteins_no_append[i]),
                set(true_possible_proteins_no_append[i]),
            )



from setuptools import setup
import glob

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    setup_requires=["pbr>=1.8", "setuptools>=17.1"],
    pbr=True,
    scripts=glob.glob("scripts/*.py"),
    name="pyproteininference",
    version="1.0.0",
    url="https://github.com/thinkle12/pyproteininference",
    license="Apache-2",
    author="Trent Hinkle",
    author_email="hinklet@gene.com",
    description="Python Package for running protein inference algorithms on tab-formatted tandem MS/MS search results.",
    keywords=['protein inference', 'proteomics', 'mass spectrometry'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
)
